import abc
import logging
import re
import typing

from hpotk.model import TermId, Term, MinimalTerm, Synonym, SynonymType, SynonymCategory, Definition
from ._model import Node, Meta, SynonymPropertyValue

logger = logging.getLogger(__name__)

MINIMAL_TERM = typing.TypeVar('MINIMAL_TERM', bound=MinimalTerm)

OBO_PURL_PT = re.compile(r'^http://purl\.obolibrary\.org/obo/(?P<value>.+)$')
HP_VAL_PT = re.compile(r'^hp(.*)#(?P<value>.+)$')
ORCID_PT = re.compile(r'.*orcid\.org/(?P<orcid>\d{4}-\d{4}-\d{4}-\d{4})$')


def create_alt_term_ids(node: Node) -> typing.List[TermId]:
    alt_term_ids = []
    if node.meta:
        for bpv in node.meta.basic_property_values:
            if bpv.pred is not None \
                    and bpv.val is not None \
                    and bpv.pred.endswith('#hasAlternativeId'):
                alt_term_ids.append(TermId.from_curie(bpv.val))
    return alt_term_ids


def create_synonyms(meta: Meta) -> typing.Optional[typing.List[Synonym]]:
    if len(meta.synonyms) == 0:
        return None
    else:
        return [parse_synonym(s) for s in meta.synonyms]


def parse_synonym(spv: SynonymPropertyValue) -> Synonym:
    synonym_category: typing.Optional[SynonymCategory] = parse_synonym_category(spv.pred)
    synonym_type: typing.Optional[SynonymType] = parse_synonym_type(spv.synonym_type)
    if len(spv.xrefs) != 0:
        xrefs = []
        for xref in spv.xrefs:
            parsed = parse_synonym_xref(xref)
            if parsed is not None:
                xrefs.append(parsed)
        xrefs = list(xrefs) if len(xrefs) != 0 else None  # shrink to fit
    else:
        xrefs = None

    return Synonym(name=spv.val, synonym_category=synonym_category, synonym_type=synonym_type, xrefs=xrefs)


def parse_synonym_category(synonym_category: str) -> typing.Optional[SynonymCategory]:
    if synonym_category == 'hasRelatedSynonym':
        return SynonymCategory.RELATED
    elif synonym_category == 'hasExactSynonym':
        return SynonymCategory.EXACT
    elif synonym_category == 'hasBroadSynonym':
        return SynonymCategory.BROAD
    elif synonym_category == 'hasNarrowSynonym':
        return SynonymCategory.NARROW
    else:
        logger.debug(f"Unknown synonym category {synonym_category}")
        return None


def parse_synonym_type(synonym_type: str) -> typing.Optional[SynonymType]:
    if synonym_type is None or len(synonym_type) == 0:
        return None
    hp_obo_matcher = OBO_PURL_PT.match(synonym_type)
    if hp_obo_matcher:
        value = hp_obo_matcher.group('value')
        hp_matcher = HP_VAL_PT.match(value)
        if hp_matcher:
            value = hp_matcher.group('value')
            if value in ('layperson', 'layperson term'):
                return SynonymType.LAYPERSON_TERM
            elif value == 'abbreviation':
                return SynonymType.ABBREVIATION
            elif value == 'uk_spelling':
                return SynonymType.UK_SPELLING
            elif value == 'obsolete_synonym':
                return SynonymType.OBSOLETE_SYNONYM
            elif value == 'plural_form':
                return SynonymType.PLURAL_FORM
        else:
            if value in ('HP_0034334', 'allelic_requirement'):
                return SynonymType.ALLELIC_REQUIREMENT

    logger.debug(f"Unknown synonym type {synonym_type}")
    return None


def parse_synonym_xref(xref) -> typing.Optional[TermId]:
    orcid_matcher = ORCID_PT.match(xref)
    if orcid_matcher:
        return TermId.from_curie(f'ORCID:{orcid_matcher.group("orcid")}')
    else:
        try:
            # TODO: this can contain many things. Investigate..
            return TermId.from_curie(xref)
        except ValueError:
            logger.debug(f'Unable to create a synonym xref from {xref}')
            return None


def create_xrefs(meta: Meta) -> typing.Optional[typing.List[TermId]]:
    if len(meta.xrefs) == 0:
        return None
    else:
        # TODO: Expecting that all xrefs are CURIES may be a bit too naive. Investigate..
        return [TermId.from_curie(xref.val) for xref in meta.xrefs]


class ObographsTermFactory(typing.Generic[MINIMAL_TERM], metaclass=abc.ABCMeta):
    """
    Term factory turns `TermId` and obographs `Node` into an ontology term.
    """

    @abc.abstractmethod
    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[MINIMAL_TERM]:
        """
        Create `MinimalTerm` or a more specific instance for `TermId` and `Node`

        The term may not be created at the discretion of the factory, in which case `None` is returned.
        """
        pass


class MinimalTermFactory(ObographsTermFactory[MinimalTerm]):

    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[MinimalTerm]:
        is_obsolete = node.meta is not None and node.meta.is_deprecated
        alt_term_ids = create_alt_term_ids(node)
        return MinimalTerm.create_minimal_term(term_id, node.lbl, alt_term_ids, is_obsolete)


class TermFactory(ObographsTermFactory[Term]):

    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[Term]:
        if node.meta:
            if node.meta.definition is not None:
                d = node.meta.definition.val
                xrefs = node.meta.definition.xrefs
                definition = Definition(d, xrefs)
            else:
                definition = None
            comment = ', '.join(node.meta.comments) if len(node.meta.comments) > 0 else None
            alt_term_ids = create_alt_term_ids(node)
            synonyms = create_synonyms(node.meta)
            xrefs = create_xrefs(node.meta)

            return Term.create_term(term_id, name=node.lbl, alt_term_ids=alt_term_ids,
                                    is_obsolete=node.meta.is_deprecated, definition=definition, comment=comment,
                                    synonyms=synonyms, xrefs=xrefs)
        else:
            return Term.create_term(term_id, name=node.lbl, alt_term_ids=[],
                                    is_obsolete=False, definition=None, comment=None,
                                    synonyms=None, xrefs=None)



import os
from pathlib import Path

import pytest

import hpotk


@pytest.mark.skip('Needs an internet connection')
class TestGitHubOntologyStore:

    def test_load_minimal_hpo(
            self,
            tmp_path: Path,
    ):
        assert len(os.listdir(tmp_path)) == 0  # We start with the clean slate.

        store = hpotk.configure_ontology_store(store_dir=str(tmp_path))
        assert len(os.listdir(tmp_path)) == 0  # Creating the store does nothing.

        release = 'v2024-03-06'
        hpo = store.load_minimal_hpo(release=release)

        assert isinstance(hpo, hpotk.MinimalOntology)
        assert hpo.version == release[1:]

        fpath_expected = os.path.join(tmp_path, 'HP', f'hp.{release}.json')
        assert os.path.isfile(fpath_expected)

    def test_load_minimal_hpo__invalid_release(
            self,
            tmp_path: Path,
    ):
        store = hpotk.configure_ontology_store(store_dir=str(tmp_path))

        release = 'v3400-12-31'
        with pytest.raises(ValueError) as e:
            store.load_minimal_hpo(release=release)

        assert e.value.args[0] == f'Could not find {release} on GitHub'



import typing

import numpy as np

import hpotk
from ._api import IndexedOntologyGraph, NODE


class StaticCsrArray:

    def __init__(self, indptr: typing.Sequence[int],
                 data: typing.Sequence[int]):
        self._indptr = np.array(indptr)
        self._data = np.array(data)
        self._max_row = len(self._indptr)

    def outgoing_nodes(self, row: int) -> typing.Sequence[int]:
        if row < 0 or self._max_row < row:
            raise ValueError(f'No such row {row}')

        start = self._indptr[row]
        end = self._indptr[row + 1]

        # TODO: maybe we can test if start==end and return a tuple if yes
        return self._data[start: end]


class CsrData:

    def __init__(self, children: StaticCsrArray,
                 parents: StaticCsrArray):
        self._children = children
        self._parents = parents

    @property
    def children(self) -> StaticCsrArray:
        return self._children

    @property
    def parents(self) -> StaticCsrArray:
        return self._parents


class CsrIndexedOntologyGraph(IndexedOntologyGraph):

    def __init__(self, root: int,
                 nodes: typing.Sequence[NODE],
                 csr_data: CsrData):
        self._root = hpotk.util.validate_instance(root, int, 'root')
        self._nodes = np.array(nodes)
        self._node_to_idx = {node: i for i, node in enumerate(self._nodes)}
        self._csr_data = hpotk.util.validate_instance(csr_data, CsrData, 'csr_data')

    @property
    def root_idx(self) -> int:
        return self._root

    def get_children_idx(self, source: int) -> typing.Sequence[int]:
        return self._csr_data.children.outgoing_nodes(source)

    def get_descendant_idx(self, source: int) -> typing.Iterator[int]:
        return self._traverse_graph(source, self.get_children_idx)

    def get_parents_idx(self, source: int) -> typing.Sequence[int]:
        return self._csr_data.parents.outgoing_nodes(source)

    def get_ancestor_idx(self, source: int) -> typing.Iterator[int]:
        return self._traverse_graph(source, self.get_parents_idx)

    def idx_to_node(self, idx: int) -> NODE:
        return self._nodes[idx]

    def node_to_idx(self, node: NODE) -> typing.Optional[int]:
        try:
            return self._node_to_idx[node]
        except KeyError:
            return None

    def __iter__(self) -> typing.Iterator[NODE]:
        return iter(self._nodes)

    @staticmethod
    def _traverse_graph(source: int,
                        supplier: typing.Callable[[int], typing.Sequence[int]]) -> typing.Generator[int, None, None]:
        seen = set()
        buffer = []

        for idx in supplier(source):
            seen.add(idx)
            buffer.append(idx)

        while buffer:
            current = buffer.pop()
            for idx in supplier(current):
                if idx not in seen:
                    seen.add(idx)
                    buffer.append(idx)

            yield current

    def __repr__(self):
        return f"CsrIndexedOntologyGraph(root={self.idx_to_node(self._root)}, n_nodes={len(self._nodes)})"



import typing
import abc
import enum

from collections import namedtuple

from hpotk.model import Identified, TermId
from ._util import SimpleFeature, map_to_stateful_feature


class ValidationLevel(enum.Enum):
    """
    An enumeration of the validation levels.
    """

    WARNING = 1
    """
    Validation warning represents a potential issue that can be fixed. 
    """

    ERROR = 2
    """
    Validation error represents an issue that MUST be fixed
    """


ValidationResult = namedtuple('ValidationResult', field_names=['level', 'category', 'message'])
"""
A tuple of :class:`ValidationLevel`, a validation `category` string, and human-centric `message`. 
"""


class ValidationResults:
    """
    A container for results of a single validation run.

    :param results:
    """

    def __init__(self, results: typing.Sequence[ValidationResult]):
        self._results = results
        
    @property
    def results(self) -> typing.Sequence[ValidationResult]:
        """
        Get a sequence of validation results.
        """
        return self._results

    def is_ok(self) -> bool:
        """
        Test if the validation run found no errors.

        :return: `True` if no errors were found or `False` otherwise.
        """
        return len(self._results) == 0

    def __str__(self):
        return f"ValidationResults(is_ok={self.is_ok()}, n_results={len(self._results)})"

    def __repr__(self) -> str:
        return f"ValidationResults(results={[self._results]})"


class RuleValidator(metaclass=abc.ABCMeta):
    """
    `RuleValidator` checks if a sequence of :class:`hpotk.model.Identified` or :class:`hpotk.model.TermId` items meet
    the validation requirements.
    The observation status is included in case the items extend :class:`hpotk.model.ObservableFeature`.

    The validators can check each item individually or as a collection, for instance,
    to discover violation of the annotation propagation rule, etc.

    The issues are returned as :class:`ValidationResults`.
    """

    @abc.abstractmethod
    def validate(self, items: typing.Sequence[typing.Union[Identified, TermId]]) -> ValidationResults:
        """
        Validate the sequence of term IDs or items that have an identifier.

        :param items: the items to validate.
        :return: results packaged in :class:`ValidationResults` container.
        """
        pass

    @staticmethod
    def _extract_stateful_feature(item: typing.Union[Identified, TermId]) -> SimpleFeature:
        return map_to_stateful_feature(item)


class ValidationRunner:
    """
    The runner applies a sequence of rule validators on items and packs the results
    into :class:`ValidationResults`.

    :param validators: an iterable of :class:`RuleValidator`\ s to apply.
    """

    def __init__(self, validators: typing.Iterable[RuleValidator]):
        self._validators = tuple(validators)

    def validate_all(self, items: typing.Sequence[typing.Union[Identified, TermId]]) -> ValidationResults:
        """
        Validate the `items` with all rules.

        :return: the results packed into :class:`ValidationResults`.
        """
        overall = []
        for validator in self._validators:
            results = validator.validate(items)
            for result in results.results:
                overall.append(result)

        return ValidationResults(overall)



import json
import unittest

from ._load import extract_ontology_version


class TestLoad(unittest.TestCase):

    def test_parse_ontology_version(self):
        meta = """
        {
          "version": "http://purl.obolibrary.org/obo/hp/releases/2022-10-05/hp.json"
        }
        """
        meta = json.loads(meta)
        version = extract_ontology_version(meta)
        self.assertEqual(version, '2022-10-05')

    def test_parse_ontology_version_from_bpv(self):
        meta = """
        {
          "basicPropertyValues": [
            {
              "pred": "http://www.w3.org/2002/07/owl#versionInfo",
              "val": "2022-10-05"
            }
          ]
        }
        """
        meta = json.loads(meta)
        version = extract_ontology_version(meta)
        self.assertEqual(version, '2022-10-05')



"""
The `hpotk.annotations` module provides classes for working with HPO annotation data that is available for download
from `HPO release data <https://hpo.jax.org/app/data/annotations>`_.

The module contains data classes to model the annotation data. Most notable classes include :class:`HpoDiseases`,
a container of diseases, and :class:`HpoDisease` a representation of the disease data.

The :mod:`hpotk.annotations.load` module contains code for loading the annotations into from HPO annotations format.
"""

from ._api import AnnotatedItem, AnnotatedItemContainer, ANNOTATED_ITEM, ANNOTATION
from ._base import EvidenceCode, Sex, AnnotationReference
from ._base import HpoDiseaseAnnotation, HpoDisease, HpoDiseases

from ._simple import SimpleHpoDiseaseAnnotation, SimpleHpoDisease, SimpleHpoDiseases

from . import load

__all__ = [
    'HpoDiseases', 'HpoDisease', 'HpoDiseaseAnnotation',
    'AnnotatedItem', 'AnnotatedItemContainer', 'ANNOTATED_ITEM', 'ANNOTATION',
    'EvidenceCode', 'Sex', 'AnnotationReference',
    'SimpleHpoDiseaseAnnotation', 'SimpleHpoDisease', 'SimpleHpoDiseases'
]



import hpotk


class TestTerms:
    """
    We only load the ontology once, and we test the properties of the loaded data.
    """

    def test_term_properties(self, toy_hpo: hpotk.Ontology):
        # Test properties of a Term
        term = toy_hpo.get_term('HP:0001626')

        assert term.identifier.value == 'HP:0001626'
        assert term.name == 'Abnormality of the cardiovascular system'
        definition = term.definition
        assert definition.definition == 'Any abnormality of the cardiovascular system.'
        assert definition.xrefs == ('HPO:probinson',)
        assert term.comment == 'The cardiovascular system consists of the heart, vasculature, and the lymphatic system.'
        assert not term.is_obsolete
        assert term.alt_term_ids == (hpotk.TermId.from_curie('HP:0003116'),)

        synonyms = term.synonyms
        assert len(synonyms) == 3

        one = synonyms[0]
        assert one.name == 'Cardiovascular disease'
        assert one.category == hpotk.model.SynonymCategory.RELATED
        assert one.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert one.xrefs is None

        two = synonyms[1]
        assert two.name == 'Cardiovascular abnormality'
        assert two.category == hpotk.model.SynonymCategory.EXACT
        assert two.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert two.xrefs is None

        three = synonyms[2]
        assert three.name == 'Abnormality of the cardiovascular system'
        assert three.category == hpotk.model.SynonymCategory.EXACT
        assert three.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert three.xrefs is None

        assert term.xrefs == tuple(
            hpotk.TermId.from_curie(curie) for curie in
            ('UMLS:C0243050', 'UMLS:C0007222', 'MSH:D018376', 'SNOMEDCT_US:49601007', 'MSH:D002318')
        )

    def test_synonym_properties(self, toy_hpo: hpotk.Ontology):
        term = toy_hpo.get_term('HP:0001627')

        synonym = term.synonyms[7]

        assert synonym.name == 'Abnormally shaped heart'
        assert synonym.category == hpotk.model.SynonymCategory.EXACT
        assert synonym.synonym_type == hpotk.model.SynonymType.LAYPERSON_TERM
        assert synonym.xrefs == [hpotk.TermId.from_curie('ORCID:0000-0001-5208-3432')]



from ._api import OntologyGraph, IndexedOntologyGraph, GraphAware, OWL_THING
from ._csr_graph import SimpleCsrOntologyGraph  # REMOVE(v1.0.0)
from ._factory import CsrGraphFactory  # REMOVE(v1.0.0)
from ._factory import GraphFactory, IncrementalCsrGraphFactory, CsrIndexedGraphFactory

__all__ = [
    'OntologyGraph', 'IndexedOntologyGraph',
    'GraphFactory', 'IncrementalCsrGraphFactory', 'CsrIndexedGraphFactory',
    'GraphAware',
]



import os
import platform
import re
import typing
from pathlib import Path

from ._api import OntologyStore
from ._github import GitHubOntologyStore


def configure_ontology_store(
        store_dir: typing.Optional[str] = None,
) -> OntologyStore:
    """
    Configure and create the default ontology store.

    :param: a `str` pointing to an existing directory for caching the ontology files
      or `None` if the platform-specific default folder should be used.
    :returns: an :class:`OntologyStore`.
    :raises: `ValueError` if something goes wrong.
    """
    if store_dir is None:
        store_dir = get_default_ontology_store_dir()
    else:
        if not os.path.isdir(store_dir):
            raise ValueError(f'`store_dir` must point to an existing directory')
    return GitHubOntologyStore(store_dir=store_dir)


def get_default_ontology_store_dir() -> str:
    """
    Get a platform specific absolute path to the data directory.

    The data directory points to `$HOME/.hpo-toolkit` on UNIX and `$HOME/hpo-toolkit` on Windows.
    The folder is created if it does not exist.
    """
    ps = platform.system()

    if re.match('(linux)|(darwin)', ps, re.IGNORECASE):
        store_dir_name = '.hpo-toolkit'
    elif re.match('windows', ps, re.IGNORECASE):
        store_dir_name = 'hpo-toolkit'
    else:
        raise ValueError(f'Unsupported platform {ps}')

    dir_name = os.path.join(Path.home(), store_dir_name)

    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)

    return dir_name



import abc
import typing

from hpotk.model import Identified, TermId
from hpotk.ontology import MinimalOntology
from hpotk.constants.hpo.base import PHENOTYPIC_ABNORMALITY
from hpotk.util import validate_instance

from ._model import ValidationResult, ValidationResults, ValidationLevel, RuleValidator
from ._util import SimpleFeature

T = typing.TypeVar('T', SimpleFeature, TermId)


class BaseOntologyRuleValidator(RuleValidator, metaclass=abc.ABCMeta):

    def __init__(self, hpo: MinimalOntology):
        self._hpo = validate_instance(hpo, MinimalOntology, 'hpo')

    def _primary_term_id(self, feature: T) -> typing.Optional[T]:
        """
        Update the term ID of the `feature` to the primary term ID in case the current term ID is obsolete.
        """
        if isinstance(feature, SimpleFeature):
            current_id = self._hpo.get_term(feature.identifier)
            if current_id is None:
                return None
            if current_id.identifier != feature.identifier:
                # Update to the primary term ID.
                feature.identifier = current_id.identifier
            return feature
        elif isinstance(feature, TermId):
            current = self._hpo.get_term(feature)
            return None if current is None else current.identifier
        else:
            raise ValueError(f'feature must be either a `TermId` or `SimpleFeature` but was {type(feature)}')


class AnnotationPropagationValidator(BaseOntologyRuleValidator):
    """
    Validator to check that a sequence of terms, when interpreted together, does not violate the annotation propagation
    rule.

    More formally, the terms must not satisfy the following:

    * terms must not contain a present term and its present or excluded ancestor
    * terms must not contain an excluded term and its excluded ancestor

    Violation of annotation propagation rule produces an :class:`ValidationLevel.ERROR`.

    The validator replaces obsolete term IDs with the current term IDs before performing the validation.

    :param ontology: HPO represented as :class:`hpotk.ontology.MinimalOntology`.
    """

    def __init__(self, ontology: MinimalOntology):
        super().__init__(ontology)

    def validate(self, items: typing.Sequence[typing.Union[Identified, TermId]]) -> ValidationResults:
        stateful_features: typing.Collection[SimpleFeature] = {
            self._primary_term_id(self._extract_stateful_feature(item))
            for item in items
            if item is not None
        }
        results = []
        for feature in stateful_features:
            if feature.is_present:
                # A present feature cannot coexist with a present or excluded ancestor.
                for anc in self._hpo.graph.get_ancestors(feature):
                    if any(anc == sf.identifier for sf in stateful_features):
                        current_term = self._hpo.get_term(feature.identifier)
                        term = self._hpo.get_term(anc)
                        results.append(
                            ValidationResult(level=ValidationLevel.ERROR,
                                             category='annotation_propagation',
                                             message=f'Terms should not contain both present '
                                                     f'{current_term.name} [{current_term.identifier.value}] '
                                                     f'and its present or excluded ancestor '
                                                     f'{term.name} [{term.identifier.value}]'))
            else:
                # An excluded feature cannot coexist with an excluded ancestor.
                for anc in self._hpo.graph.get_ancestors(feature):
                    # We only check excluded ancestors here since presence of an ancestor is allowed
                    # for an excluded feature.
                    if any(anc == sf.identifier for sf in stateful_features if sf.is_excluded):
                        current_term = self._hpo.get_term(feature.identifier)
                        term = self._hpo.get_term(anc)
                        results.append(
                            ValidationResult(level=ValidationLevel.ERROR,
                                             category='annotation_propagation',
                                             message=f'Terms should not contain both excluded '
                                                     f'{current_term.name} [{current_term.identifier.value}] '
                                                     f'and its present or excluded ancestor '
                                                     f'{term.name} [{term.identifier.value}]'))

        return ValidationResults(results)


class PhenotypicAbnormalityValidator(BaseOntologyRuleValidator):
    """
    Validator for checking that the term is a phenotypic abnormality
    (a descendant of `Phenotypic abnormality <https://hpo.jax.org/app/browse/term/HP:0000118>`_ [HP:0000118]).

    Presence of a term that is not a descendant of Phenotypic abnormality is a :class:`ValidationLevel.WARNING`.

    The validator replaces obsolete term IDs with the current term IDs before performing the validation.

    :param ontology: HPO represented as :class:`hpotk.ontology.MinimalOntology`.
    """

    def __init__(self, ontology: MinimalOntology):
        super().__init__(ontology)

    def validate(self, items: typing.Sequence[typing.Union[Identified, TermId]]) -> ValidationResults:
        results = []
        for item in items:
            primary = self._primary_term_id(self._extract_stateful_feature(item))
            if primary is None:
                # Unable to get the primary term ID. Handling items with obsolete IDs is not the responsibility
                # of this validator
                continue
            if not any(PHENOTYPIC_ABNORMALITY == anc for anc in self._hpo.graph.get_ancestors(primary)):
                item = self._hpo.get_term(primary.identifier)
                results.append(
                    ValidationResult(
                        level=ValidationLevel.WARNING,
                        category='phenotypic_abnormality_descendant',
                        message=f'{item.name} [{item.identifier.value}] '
                                f'is not a descendant of Phenotypic abnormality [{PHENOTYPIC_ABNORMALITY.value}]'
                    )
                )

        return ValidationResults(results)


class ObsoleteTermIdsValidator(BaseOntologyRuleValidator):
    """
    `ObsoleteTermIdsValidator` points out usage of obsolete term ids in `items`.

    Presence of an obsolete term ID is a :class:`ValidationLevel.WARNING`.

    :param ontology: HPO represented as :class:`hpotk.ontology.MinimalOntology`.
    """

    def __init__(self, ontology: MinimalOntology):
        super().__init__(ontology)

    def validate(self, items: typing.Sequence[typing.Union[Identified, TermId]]) -> ValidationResults:
        results = []
        for item in items:
            sf = self._extract_stateful_feature(item)
            current = sf.identifier  # cache the ID since _primary_term_id can update in place.
            primary = self._primary_term_id(sf)
            if primary.identifier != current:
                current_term = self._hpo.get_term(sf.identifier)
                results.append(
                    ValidationResult(
                        level=ValidationLevel.WARNING,
                        category='obsolete_term_id_is_used',
                        message=f'Using the obsolete {current.value} instead of {primary.identifier.value} '
                                f'for {current_term.name}'
                    )
                )

        return ValidationResults(results)



import math
import typing
from collections import Counter

from hpotk.model import TermId
from hpotk.annotations import AnnotatedItemContainer
from hpotk.ontology import MinimalOntology
from hpotk.util import validate_instance
from ._model import SimpleAnnotationIcContainer, AnnotationIcContainer


def calculate_ic_for_annotated_items(items: AnnotatedItemContainer,
                                     ontology: MinimalOntology,
                                     base: typing.Optional[float] = None,
                                     module_root: typing.Optional[TermId] = None,
                                     use_pseudocount: bool = False) -> AnnotationIcContainer:
    """
    Calculate information content (IC) for each :class:`TermId` based on a collection of annotated `items`.

    The calculation can be done for an ontology module - only the descendants of the provided `module_root`
    will be included in the analysis. If `assume_annotated` is `True`, then the count of all ontology/module terms
    is set to at least 1, even for those terms that do not annotate the `items`.

    :param items: a collection of :class:`hpotk.annotations.AnnotatedItem`\ s
    :param ontology: ontology with concepts used to annotate the `items`
    :param base: information content base or `None` for *e*
                 (produces IC in `nats <https://en.wikipedia.org/wiki/Nat_(unit)>`_)
    :param module_root: the root of the ontology module to calculate the IC for.
    :param use_pseudocount: assume that each ontology term annotates at least one of the `items`.

    :return: a container with mappings from :class:`TermId` to information content in nats, bits, or else,
             depending on the `base` value
    """
    ontology = validate_instance(ontology, MinimalOntology, 'ontology')

    graph = ontology.graph
    term_id_count: Counter[TermId] = Counter()
    module_term_ids: typing.Optional[typing.Set[TermId]] = None \
        if module_root is None \
        else set(graph.get_descendants(module_root, include_source=True))

    for item in items:
        for annotation in item.annotations:
            if annotation.is_present:
                if module_root is not None and annotation.identifier not in module_term_ids:
                    # annotation is not from the target module.
                    continue

                for ancestor in graph.get_ancestors(annotation.identifier, include_source=True):
                    if module_term_ids is None:
                        # Not doing module
                        term_id_count[ancestor] += 1
                    elif ancestor in module_term_ids:
                        # Doing module and the ancestor is from the module
                        term_id_count[ancestor] += 1

    if use_pseudocount:
        # Set the count of all primary term IDs to at least one but DO NOT increment the count of the ancestor
        # that already count>=1 .
        # Note, in the HPO case, this will set count of non-phenotypic abnormalities (e.g. Clinical modifier)
        # to 1 as well.
        corpus = map(lambda t: t.identifier, ontology.terms) \
            if module_root is None \
            else module_term_ids
        for term_id in corpus:
            if term_id not in term_id_count:
                term_id_count[term_id] = 1

    log_func = math.log if base is None else lambda c: math.log(c, base)

    population_count = term_id_count[graph.root] if module_root is None else term_id_count[module_root]

    data = {term_id: -log_func(count / population_count) for term_id, count in term_id_count.items()}
    metadata = {'annotated_items_version': items.version, 'ontology_version': ontology.version}
    return SimpleAnnotationIcContainer(data, metadata=metadata)



import abc
import typing

from hpotk.annotations import HpoDiseases


class HpoDiseaseLoader(metaclass=abc.ABCMeta):
    """
    Loader for reading data into :class:`HpoDiseases` container.
    """

    @abc.abstractmethod
    def load(self, file: typing.Union[typing.IO, str]) -> HpoDiseases:
        pass



"""
HPO toolkit is a library for working with Human Phenotype Ontology and the HPO annotation data.
"""

__version__ = "0.5.0"

from . import algorithm
from . import annotations
from . import constants
from . import graph
from . import model
from . import ontology
from . import util
from . import validate

from .graph import OntologyGraph, GraphAware
from .model import TermId, Term, MinimalTerm, Synonym, SynonymType, SynonymCategory
from .ontology import Ontology, MinimalOntology

from .ontology.load.obographs import load_minimal_ontology, load_ontology
from .util.store import OntologyType, OntologyStore, configure_ontology_store



from hpotk.model import TermId

# Descendents of Severity `HP:0012824`.

SEVERITY: TermId = TermId.from_curie("HP:0012824")

BORDERLINE: TermId = TermId.from_curie("HP:0012827")
MILD: TermId = TermId.from_curie("HP:0012825")
MODERATE: TermId = TermId.from_curie("HP:0012826")
SEVERE: TermId = TermId.from_curie("HP:0012828")
PROFOUND: TermId = TermId.from_curie("HP:0012829")



from . import sort  # TODO: probably not necessary

from ._io import looks_like_url, looks_gzipped
from ._io import open_text_io_handle, open_text_io_handle_for_reading, open_text_io_handle_for_writing
from ._log import setup_logging, DEFAULT_LOG_FMT
from ._validate import validate_instance, validate_optional_instance

__all__ = [
    'looks_like_url', 'looks_gzipped',
    'open_text_io_handle_for_reading', 'open_text_io_handle_for_writing',
    'setup_logging', 'DEFAULT_LOG_FMT',
    'validate_instance', 'validate_optional_instance'
]



"""
The `hpotk.util.sort` package sorts term IDs in a meaningful way. See :ref:`sort-term-ids` section for more info.
"""

from ._api import TermIdSorting
from ._hierarchical import HierarchicalSimilaritySorting, HierarchicalIcTermIdSorting, HierarchicalEdgeTermIdSorting

__all__ = [
    'TermIdSorting',
    'HierarchicalEdgeTermIdSorting', 'HierarchicalIcTermIdSorting',
    'HierarchicalSimilaritySorting',
]



import hpotk
import abc
import enum
import typing

from ._base import Identified, Named
from ._term_id import TermId


class MinimalTerm(Identified, Named, metaclass=abc.ABCMeta):
    """
    `MinimalTerm` is a data object with the minimal useful information about an ontology concept.

    Each term has:

    * `identifier` - a :class:`TermId` of the term (thanks to inheriting from :class:`Identified`)
    * `name` - a human-friendly name of the term (thanks to inheriting from :class:`Named`)
    * `alt_term_ids` - a sequence of *alternate identifiers* (IDs of obsolete terms that should be replaced by this term)
    * `is_obsolete` - the *obsoletion status*

    Most of the time, you should get terms from an :class:`hpotk.ontology.MinimalOntology`. However, `MinimalTerm` can
    also be created from scratch using :func:`create_minimal_term` if you must do that from whatever reason.
    """

    @staticmethod
    def create_minimal_term(term_id: typing.Union[TermId, str],
                            name: str,
                            alt_term_ids: typing.Iterable[typing.Union[TermId, str]],
                            is_obsolete: bool):
        """
        Create `MinimalTerm` from the components.

        .. doctest::

          >>> seizure = MinimalTerm.create_minimal_term(term_id='HP:0001250', name='Seizure',
          ...                                           alt_term_ids=('HP:0002279', 'HP:0002391'),
          ...                                           is_obsolete=False)


        :param term_id: a `TermId` or a CURIE `str` (e.g. 'HP:0001250').
        :param name: term name (e.g. Seizure) .
        :param alt_term_ids: an iterable with term IDs that represent the alternative IDs of the term.
        :param is_obsolete: `True` if the `MinimalTerm` has been obsoleted, or `False` otherwise.
        :return: the created term.
        """
        return DefaultMinimalTerm(term_id, name, alt_term_ids, is_obsolete)

    @property
    @abc.abstractmethod
    def alt_term_ids(self) -> typing.Sequence[TermId]:
        """
        Get a sequence of identifiers of the ontology concepts that were obsoleted and should be replaced by
        the concept represented by this `Term`.
        """
        pass

    @property
    def is_current(self) -> bool:
        """
        Return `True` if the term is current (*not* obsolete) and `False` otherwise.
        """
        return not self.is_obsolete

    @property
    @abc.abstractmethod
    def is_obsolete(self) -> bool:
        """
        Return `True` if the term is obsolete (*not* current) and `False` otherwise.
        """
        pass

    def __eq__(self, other):
        return isinstance(other, MinimalTerm) \
            and self.identifier == other.identifier \
            and self.name == other.name \
            and self.alt_term_ids == other.alt_term_ids \
            and self.is_obsolete == other.is_obsolete

    def __str__(self):
        return f'MinimalTerm(' \
               f'identifier="{self.identifier}", ' \
               f'name="{self.name}", ' \
               f'alt_term_ids={self.alt_term_ids}, ' \
               f'is_obsolete={self.is_obsolete})'


class SynonymCategory(enum.Enum):
    """
    An enumeration of the synonym categories.
    """

    EXACT = enum.auto()
    RELATED = enum.auto()
    BROAD = enum.auto()
    NARROW = enum.auto()


class SynonymType(enum.Enum):
    """
    An enumeration of the synonym types, as provided by Obographs.
    """

    LAYPERSON_TERM = enum.auto()
    ABBREVIATION = enum.auto()
    UK_SPELLING = enum.auto()
    OBSOLETE_SYNONYM = enum.auto()
    PLURAL_FORM = enum.auto()
    ALLELIC_REQUIREMENT = enum.auto()

    # We may not need these, unless we support ECTO
    # IUPAC_NAME = enum.auto()
    # INN = enum.auto()
    # BRAND_NAME = enum.auto()
    # IN_PART = enum.auto()
    # SYNONYM = enum.auto()
    # BLAST_NAME = enum.auto()
    # GENBANK_COMMON_NAME = enum.auto()
    # COMMON_NAME = enum.auto()

    def is_obsolete(self) -> bool:
        """
        Returns `True` if the synonym is obsolete (not current) and `False` otherwise.
        """
        return self == SynonymType.OBSOLETE_SYNONYM

    def is_current(self) -> bool:
        """
        Returns `True` if the synonym is current (not obsolete) and `False` otherwise.
        """
        return not self.is_obsolete()


class Synonym(Named):
    """
    `Synonym` represents the information regarding a synonym of an ontology concept.
    """

    def __init__(self, name: str,
                 synonym_category: typing.Optional[SynonymCategory] = None,
                 synonym_type: typing.Optional[SynonymType] = None,
                 xrefs: typing.Optional[typing.Sequence[TermId]] = None):
        self._name = name
        self._scat = synonym_category
        self._stype = synonym_type
        self._xrefs = xrefs

    @property
    def name(self) -> str:
        """
        Get the name of the synonym
        """
        return self._name

    @property
    def category(self) -> typing.Optional[SynonymCategory]:
        """
        Get the synonym category - an instance of :class:`SynonymCategory` or ``None``.
        """
        return self._scat

    @property
    def synonym_type(self) -> typing.Optional[SynonymType]:
        """
        Get synonym type - an instance of :class:`SynonymType` or ``None``.
        """
        return self._stype

    @property
    def xrefs(self) -> typing.Optional[typing.Sequence[TermId]]:
        """
        Get a sequence of identifiers of the cross-references of the ontology concept or ``None`` if there are none.
        """
        return self._xrefs

    def __eq__(self, other):
        return isinstance(other, Synonym) \
          and self.name == other.name \
          and self.category == other.category \
          and self.synonym_type == other.synonym_type \
          and self.xrefs == other.xrefs

    def __str__(self):
        return f'Synonym(' \
               f'name="{self.name}", ' \
               f'category={self.category}, ' \
               f'synonym_type={self.synonym_type}, ' \
               f'xrefs="{self.xrefs}")'

    def __repr__(self):
        return str(self)


class Definition:
    """
    `Definition` includes a definition and the cross-references.

    :param definition: a definition,
      e.g. *Abnormally long and slender fingers ("spider fingers").* for *Arachnodactyly*.
    :param xrefs: an iterable with definition cross-references,
      e.g. `('https://orcid.org/0000-0002-0736-9199',)` for *Arachnodactyly*.
    """

    def __init__(
            self,
            definition: str,
            xrefs: typing.Iterable[str],
    ):
        self._definition = hpotk.util.validate_instance(definition, str, 'definition')
        self._xrefs = tuple(xrefs)

    @property
    def definition(self) -> str:
        """
        Get a `str` with the term definition.

        For instance, *Abnormally long and slender fingers ("spider fingers").* for *Arachnodactyly*.
        """
        return self._definition

    @property
    def xrefs(self) -> typing.Sequence[str]:
        """
        Get definition of cross-references of a definition.

        For instance, `('https://orcid.org/0000-0002-0736-9199',)` for *Arachnodactyly*.
        """
        return self._xrefs

    def __eq__(self, other):
        return isinstance(other, Definition) \
          and self.definition == other.definition \
          and self.xrefs == other.xrefs

    def __str__(self):
        return f'Definition(' \
               f'definition="{self.definition}", ' \
               f'xrefs={self.xrefs}")'

    def __repr__(self):
        return str(self)


class Term(MinimalTerm, metaclass=abc.ABCMeta):
    """
    A comprehensive representation of an ontology concept.

    `Term` has all attributes of the :class:`MinimalTerm` plus the following:

    * `definition` - an optional definition of the term, including a comprehensive description and cross-references
    * `comment` - an optional comment
    * `synonyms` - an optional sequence of term synonyms
    * `cross-references` - an optional sequence of cross-references

    Most of the time, you should be getting terms from :class:`hpotk.ontology.Ontology`.
    However, if you absolutely must craft a term or two by hand, use :func:`create_term` function.
    """

    # TODO - add the remaining attributes from phenol's Term?

    @staticmethod
    def create_term(identifier: typing.Union[TermId, str],
                    name: str,
                    alt_term_ids: typing.Iterable[typing.Union[TermId, str]],
                    is_obsolete: bool,
                    definition: typing.Optional[typing.Union[Definition, str]],
                    comment: typing.Optional[str],
                    synonyms: typing.Optional[typing.Iterable[Synonym]],
                    xrefs: typing.Optional[typing.Iterable[TermId]]):
        """
        Create a `MinimalTerm` from the components.

        :param identifier: a `TermId` or a CURIE (e.g. 'HP:0001250').
        :param name: term name (e.g. Seizure).
        :param alt_term_ids: an iterable with term IDs that represent the alternative IDs of the term.
        :param is_obsolete: `True` if the `MinimalTerm` has been obsoleted, or `False` otherwise.
        :param definition: an optional `str` with a definition of the term or a :class:`Definition` with the full info.
        :param comment: an optional comment of the term.
        :param synonyms: an optional iterable with all synonyms of the term.
        :param xrefs: an optional iterable with all the cross-references.
        :return: the created term.
        """
        if isinstance(definition, str):
            definition = Definition(definition, ())
        return DefaultTerm(identifier, name, alt_term_ids, is_obsolete, definition, comment, synonyms, xrefs)

    @property
    @abc.abstractmethod
    def definition(self) -> typing.Optional[Definition]:
        """
        Get the definition of the ontology concept.
        """
        pass

    @property
    @abc.abstractmethod
    def comment(self) -> typing.Optional[str]:
        """
        Get the comment string of the ontology concept.
        """
        pass

    @property
    @abc.abstractmethod
    def synonyms(self) -> typing.Optional[typing.Sequence[Synonym]]:
        """
        Get a sequence of all synonyms (including obsolete) of the ontology concept or `None` if the concept
        has no synonyms.
        """
        pass

    def current_synonyms(self) -> typing.Iterable[Synonym]:
        """
        Get an iterable with *current* synonyms of the ontology concept.

        The iterable is empty if the concept has no current synonyms.
        """
        return self._synonyms_iter(lambda synonym:
                                   synonym.synonym_type is None
                                   or synonym.synonym_type.is_current())

    def obsolete_synonyms(self) -> typing.Iterable[Synonym]:
        """
        Get an iterable with *obsolete* synonyms of the ontology concept.

        The iterable is empty if the concept has no obsolete synonyms.
        """
        return self._synonyms_iter(lambda synonym:
                                   synonym.synonym_type is not None
                                   and synonym.synonym_type.is_obsolete())

    def _synonyms_iter(self, filter_f):
        if self.synonyms is None:
            return ()
        else:
            for synonym in self.synonyms:
                if filter_f(synonym):
                    yield synonym

    @property
    @abc.abstractmethod
    def xrefs(self) -> typing.Optional[typing.Sequence[TermId]]:
        """
        Get a sequence of the cross-references of the ontology concept.
        """
        pass

    def __eq__(self, other):
        return MinimalTerm.__eq__(self, other) \
            and isinstance(other, Term) \
            and self.definition == other.definition \
            and self.comment == other.comment \
            and self.synonyms == other.synonyms \
            and self.xrefs == other.xrefs

    def __str__(self):
        return f'Term(' \
               f'identifier={self.identifier}, ' \
               f'name="{self.name}", ' \
               f'definition={self.definition}, ' \
               f'comment={self.comment}, ' \
               f'synonyms={self.synonyms}, ' \
               f'xrefs={self.xrefs}, ' \
               f'is_obsolete={self.is_obsolete}, ' \
               f'alt_term_ids="{self.alt_term_ids}")'


def map_to_term_id(value: typing.Union[TermId, str]) -> TermId:
    if isinstance(value, TermId):
        return value
    elif isinstance(value, str):
        return TermId.from_curie(value)
    else:
        raise ValueError(f'Expected a `TermId` or `str` but got {type(value)}')


def validate_name(name: typing.Optional[str]) -> str:
    # Some obsolete nodes do not have labels in the Obographs format.
    # We assign an empty string.
    if name is None:
        return ''
    else:
        return hpotk.util.validate_instance(name, str, 'name')


class DefaultMinimalTerm(MinimalTerm):

    def __init__(self, identifier: typing.Union[TermId, str],
                 name: str,
                 alt_term_ids: typing.Iterable[typing.Union[TermId, str]],
                 is_obsolete: bool):
        self._id = hpotk.util.validate_instance(map_to_term_id(identifier), TermId, 'identifier')
        self._name = validate_name(name)
        self._alts = tuple(map(map_to_term_id, alt_term_ids))
        self._is_obsolete = hpotk.util.validate_instance(is_obsolete, bool, 'is_obsolete')

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def alt_term_ids(self) -> typing.Sequence[TermId]:
        return self._alts

    @property
    def is_obsolete(self) -> bool:
        return self._is_obsolete

    def __repr__(self):
        return f'DefaultMinimalTerm(' \
               f'identifier={self._id},' \
               f' name="{self._name}",' \
               f' is_obsolete={self._is_obsolete},' \
               f' alt_term_ids="{self._alts}")'


def validate_synonyms(synonyms: typing.Optional[typing.Iterable[Synonym]]):
    if synonyms is None:
        return None
    else:
        validated = []
        for i, s in enumerate(synonyms):
            validated.append(hpotk.util.validate_instance(s, Synonym, f'synonym #{i}'))
        return tuple(validated)


class DefaultTerm(DefaultMinimalTerm, Term):

    def __init__(self, identifier: typing.Union[TermId, str],
                 name: str,
                 alt_term_ids: typing.Iterable[typing.Union[TermId, str]],
                 is_obsolete: bool,
                 definition: typing.Optional[Definition],
                 comment: typing.Optional[str],
                 synonyms: typing.Optional[typing.Iterable[Synonym]],
                 xrefs: typing.Optional[typing.Iterable[typing.Union[TermId, str]]]):
        DefaultMinimalTerm.__init__(self, identifier=identifier, name=name,
                                    alt_term_ids=alt_term_ids, is_obsolete=is_obsolete)
        self._definition = hpotk.util.validate_optional_instance(definition, Definition, 'definition')
        self._comment = hpotk.util.validate_optional_instance(comment, str, 'comment')
        self._synonyms = validate_synonyms(synonyms)
        self._xrefs = tuple(map(map_to_term_id, xrefs)) if xrefs is not None else None

    @property
    def definition(self) -> typing.Optional[Definition]:
        return self._definition

    @property
    def comment(self) -> typing.Optional[str]:
        return self._comment

    @property
    def synonyms(self) -> typing.Optional[typing.Sequence[Synonym]]:
        return self._synonyms

    @property
    def xrefs(self) -> typing.Optional[typing.Sequence[TermId]]:
        return self._xrefs

    def __repr__(self):
        return f'DefaultTerm(' \
               f'identifier={self._id}, ' \
               f'name="{self._name}", ' \
               f'definition={self._definition}, ' \
               f'comment={self._comment}, ' \
               f'synonyms={self._synonyms}, ' \
               f'xrefs={self._xrefs}, ' \
               f'is_obsolete={self._is_obsolete}, ' \
               f'alt_term_ids="{self._alts}")'



import abc
import typing
import warnings

import numpy as np

from hpotk.graph import OntologyGraph, GraphAware
from hpotk.model import TermId, Identified

from ._api import TermIdSorting


def to_term_id(item: typing.Union[TermId, Identified]) -> TermId:
    """
    Private utility method for converting the input into a `TermId`.
    """
    if isinstance(item, TermId):
        return item
    elif isinstance(item, Identified):
        return item.identifier
    else:
        raise ValueError(f'Item {item} is not `TermId` or `Identified` but it is {type(item)}')


def to_ontology_graph(hpo: typing.Union[GraphAware, OntologyGraph]) -> OntologyGraph:
    if isinstance(hpo, OntologyGraph):
        return hpo
    elif isinstance(hpo, GraphAware):
        return hpo.graph
    else:
        raise ValueError(f'`hpo` must be an instance of `OntologyGraph` or `GraphAware` but was {type(hpo)}')


class Node(Identified):
    # Private API

    @staticmethod
    def merge_nodes(identifier, left, right):
        return Node(identifier, tag=False, left=left, right=right)

    @staticmethod
    def make_tagged_node(identifier: TermId):
        return Node(identifier=identifier, tag=True)

    def __init__(self, identifier: TermId,
                 tag: bool,
                 left=None,
                 right=None):
        self._id = identifier
        self._is_tagged = tag
        self._left = left
        self._right = right

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def is_tagged(self) -> bool:
        return self._is_tagged

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    def __repr__(self):
        return (f"Node(identifier={self._id}, "
                f"is_tagged={self._is_tagged}, "
                f"left={self._left}, "
                f"right={self._right})")


class SimilarityMeasure(metaclass=abc.ABCMeta):
    # Private API

    @abc.abstractmethod
    def compute_similarity(self, left: typing.Union[TermId, Identified],
                           right: typing.Union[TermId, Identified]) -> typing.Tuple[float, TermId]:
        """
        Compute similarity of two term IDs along with the most specific node.

        :param left: an ontology node.
        :param right: other ontology node.
        :return: a tuple with node similarity and the most specific node
        """
        pass


class IcSimilarityMeasure(SimilarityMeasure):
    """
    `IcSimilarityMeasure` uses information content of the most informative common ancestor as the similarity measure
    and MICA as the most specific node.
    """
    # Private API

    def __init__(self, hpo: typing.Union[OntologyGraph, GraphAware],
                 ic_source: typing.Callable[[TermId], float]):
        self._hpo = to_ontology_graph(hpo)
        self._ic_source = ic_source

    def compute_similarity(self, left: typing.Union[TermId, Identified],
                           right: typing.Union[TermId, Identified]) -> typing.Tuple[float, TermId]:

        # Find the common ancestors of `a` and `b` and find the most informative common ancestor
        # along with its information content.
        a_anc = set(self._hpo.get_ancestors(left, include_source=True))
        b_anc = set(self._hpo.get_ancestors(right, include_source=True))
        common_ancestors = a_anc.intersection(b_anc)

        return max(
            map(lambda t: (self._ic_source(t), t), common_ancestors),
            key=lambda tup: tup[0],  # Order by similarity
            default=(0., None)
        )


class EdgeSimilarityMeasure(SimilarityMeasure):
    """
    `EdgeSimilarityMeasure` uses the reciprocal of the edge distance as a similarity and most specific common ancestor
    as the most specific node.

    Note, if two paths exist between the query nodes through two separate nodes that also happen to be the most specific
    common ancestors, then one of them is chosen at random.
    """
    # Private API

    def __init__(self, hpo: typing.Union[OntologyGraph, GraphAware]):
        self._hpo = to_ontology_graph(hpo)

    def compute_similarity(self, left: typing.Union[TermId, Identified],
                           right: typing.Union[TermId, Identified]) -> typing.Tuple[float, TermId]:
        dist, node = self.calculate_edge_distance(left, right)

        sim = 1 if dist == 0 else 1 / dist

        return sim, node

    def calculate_edge_distance(self, left: TermId, right: TermId) -> typing.Tuple[int, TermId]:
        left, right = to_term_id(left), to_term_id(right)
        if left == right:
            # Distance to self is `0`.
            return 0, left

        left_dist = self._get_ancestor_distances(left)
        right_dist = self._get_ancestor_distances(right)

        return self._find_minimum_distance(left_dist, right_dist)

    def _get_ancestor_distances(self, src: TermId) -> typing.Mapping[TermId, int]:
        distances = {}
        seen = set()

        stack = [(0, src)]
        while stack:
            distance, term_id = stack.pop()

            current = distance + 1
            for parent in self._hpo.get_parents(term_id):
                if parent not in seen:
                    stack.append((current, parent))
                seen.add(parent)

            if term_id in distances:
                # We must keep the shortest distance
                distances[term_id] = min(distances[term_id], distance)
            else:
                distances[term_id] = distance

        return distances

    @staticmethod
    def _find_minimum_distance(left_dist: typing.Mapping[TermId, int],
                               right_dist: typing.Mapping[TermId, int]) -> typing.Tuple[int, TermId]:
        dist = None
        ancestor = None
        for shared in left_dist.keys() & right_dist.keys():
            current = left_dist[shared] + right_dist[shared]
            if dist is None:
                dist = current
                ancestor = shared
            else:
                if current < dist:
                    # Note: we do not update if current = dist. In result, I suspect that the algorithm is not "stable"
                    # in terms of sorting.
                    dist = current
                    ancestor = shared

        return dist, ancestor


class HierarchicalSorting(TermIdSorting, metaclass=abc.ABCMeta):
    """
    `HierarchicalSorting` implements the hierarchical clustering functionality using
    given similarity measure.
    """
    # Private API

    def __init__(self, hpo: typing.Union[OntologyGraph, GraphAware],
                 sim_measure: SimilarityMeasure,
                 epsilon: float = 5e-10):
        self._hpo = to_ontology_graph(hpo)
        self._sim_measure = sim_measure
        self._epsilon = epsilon

    def argsort(self, term_ids: typing.Sequence[typing.Union[TermId, Identified]]) -> typing.Sequence[int]:
        if len(term_ids) == 0:
            raise ValueError(f'Term ID sequence must not be empty!')

        term_ids = tuple(self._to_term_id(item) for item in term_ids)
        nodes = list(Node.make_tagged_node(tid) for tid in term_ids)

        node = self._hierarchical_cluster(nodes)

        ordered = []
        self._inorder_walk(node, ordered)

        return self._find_indices(term_ids, ordered)

    def _hierarchical_cluster(self, nodes: typing.MutableSequence[Node]) -> Node:
        while len(nodes) > 1:
            n_nodes = len(nodes)
            sims = np.zeros(shape=(n_nodes, n_nodes))
            ancestors = np.empty(shape=(n_nodes, n_nodes), dtype=object)

            for row in range(0, n_nodes):
                for col in range(row + 1, n_nodes):
                    ic, sim = self._sim_measure.compute_similarity(nodes[row], nodes[col])
                    sims[row, col] = ic
                    sims[col, row] = ic
                    ancestors[row, col] = sim
                    ancestors[col, row] = sim

            max_row, max_col = np.unravel_index(np.argmax(sims, axis=None), sims.shape)
            sim_anc = sims[max_row, max_col]

            if sim_anc <= self._epsilon:
                # sim is close to zero - should happen when there is no good match left for the remaining nodes,
                # and we need to cluster the rest kind of arbitrarily
                a = nodes.pop()
                b = nodes.pop()
                _, sim = self._sim_measure.compute_similarity(a, b)
                node = Node.merge_nodes(sim, a, b)
            else:
                # Let's cluster together the most similar term id pair.
                max_term_id = ancestors[max_row, max_col]
                a = nodes.pop(max(max_row, max_col))
                b = nodes.pop(min(max_row, max_col))
                node = Node.merge_nodes(max_term_id, a, b)

            nodes.append(node)

        return nodes[0]

    def _inorder_walk(self, node: Node,
                      collector: typing.MutableSequence[TermId]):
        # Traverse the node and collect the tagged nodes in the collector.
        if node.left is not None:
            self._inorder_walk(node.left, collector)
        if node.is_tagged:
            collector.append(node.identifier)
        if node.right is not None:
            self._inorder_walk(node.right, collector)

    @staticmethod
    def _to_term_id(item: typing.Union[TermId, Identified]) -> TermId:
        if isinstance(item, TermId):
            return item
        elif isinstance(item, Identified):
            return item.identifier
        else:
            raise ValueError(f'Item {item} is not `TermId` or `Identified` but it is {type(item)}')

    @staticmethod
    def _find_indices(source: typing.Sequence[TermId],
                      ordered: typing.Sequence[TermId]) -> typing.Sequence[int]:
        """
        Find indices that will sort `source` to the order of `ordered` sequence.
        """
        assert len(source) == len(ordered)

        return tuple(source.index(s) for s in ordered)


class HierarchicalEdgeTermIdSorting(HierarchicalSorting):
    """
    `HierarchicalEdgeTermIdSorting` uses hierarchical clustering to sort the term IDs.
    The clustering uses edge distance as a proxy to the term similarity.

    Notes:

    - Does *not* maintain order if the input is already sorted.

    :param hpo: HPO ontology graph or a graph aware instance.
    """

    def __init__(self, hpo: typing.Union[OntologyGraph, GraphAware]):
        super().__init__(hpo, EdgeSimilarityMeasure(hpo))


class HierarchicalIcTermIdSorting(HierarchicalSorting):
    """
    `HierarchicalIcTermIdSorting` uses hierarchical clustering to sort the term IDs.
    The clustering uses Resnik term similarity to assess term similarity.

    Notes:

    - Does *not* maintain order if the input is already sorted.

    :param hpo: HPO ontology graph or a graph aware instance.
    :param ic_source: a callable for getting the information content (IC) as a `float` for a term ID.
    """

    def __init__(self, hpo: typing.Union[OntologyGraph, GraphAware],
                 ic_source: typing.Callable[[TermId], float]):
        super().__init__(hpo, IcSimilarityMeasure(hpo, ic_source))

    def argsort(self, term_ids: typing.Sequence[typing.Union[TermId, Identified]]) -> typing.Sequence[int]:
        if len(term_ids) == 0:
            raise ValueError(f'Term ID sequence must not be empty!')

        term_ids = tuple(to_term_id(item) for item in term_ids)
        nodes = list(Node.make_tagged_node(tid) for tid in term_ids)

        node = self._hierarchical_cluster(nodes)

        ordered = []
        self._inorder_walk(node, ordered)

        return self._find_indices(term_ids, ordered)


class HierarchicalSimilaritySorting(HierarchicalIcTermIdSorting):
    """
    `HierarchicalSimilaritySorting` uses hierarchical clustering to sort the term IDs.
    The clustering uses Resnik term similarity to assess term similarity.

    Notes:

    - Does *not* maintain order if the input is already sorted.

    :param hpo: HPO ontology graph or a graph aware instance.
    :param ic_source: a callable for getting the information content (IC) as a `float` for a term ID.
    """

    def __init__(self, hpo: typing.Union[OntologyGraph, GraphAware],
                 ic_source: typing.Callable[[TermId], float]):
        super().__init__(hpo, ic_source)
        warnings.warn("HierarchicalSimilaritySorting was deprecated and will be remove in v1.0.0. "
                      "Use HierarchicalIcTermIdSorting instead", DeprecationWarning, stacklevel=2)



"""
The `hpotk.model` package provides data structures for working with ontology data.
"""

import typing

from ._term_id import TermId
from ._base import Identified, ObservableFeature, FrequencyAwareFeature, Named, Versioned, MetadataAware
from ._term import MinimalTerm, Term, Synonym, SynonymType, SynonymCategory, Definition

# Types
ID = typing.TypeVar('ID', bound=TermId)
CURIE_OR_TERM_ID = typing.Union[str, ID]
CURIE_OR_TERM_ID_OR_IDENTIFIED = typing.Union[str, TermId, Identified]
MINIMAL_TERM = typing.TypeVar('MINIMAL_TERM', bound=MinimalTerm)
TERM = typing.TypeVar('TERM', bound=Term)

__all__ = [
    'TermId', 'MinimalTerm', 'Term',
    'Identified', 'ObservableFeature', 'FrequencyAwareFeature', 'Named', 'Versioned', 'MetadataAware',
    'Synonym', 'SynonymType', 'SynonymCategory', 'Definition',
    'ID', 'CURIE_OR_TERM_ID', 'CURIE_OR_TERM_ID_OR_IDENTIFIED', 'MINIMAL_TERM', 'TERM'
]



import logging

DEFAULT_LOG_FMT = '%(asctime)s %(name)-20s %(levelname)-3s : %(message)s'


def setup_logging(level: int = logging.INFO,
                  log_fmt: str = DEFAULT_LOG_FMT):
    """
    Create a basic configuration for the logging library. Set up console and file handler using provided `log_fmt`.

    The `level` signifies the desired verbosity.
    Use:

    - `10` for `DEBUG`
    - `20` for `INFO`
    - `30` for `WARNING`
    - `40` for `ERROR`
    - `50` for `CRITICAL`

    :param level: the verbosity to use, `INFO` by default.
    :param log_fmt: format string for logging.
    """
    # create logger
    logger = logging.getLogger()
    logger.setLevel(level)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(level)
    # create formatter
    formatter = logging.Formatter(log_fmt)
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(ch)




import abc
import bisect
import typing
import warnings
from collections import deque

import numpy as np

from hpotk.model import TermId, Identified
from hpotk.util import validate_instance
from ._api import OntologyGraph, NODE
from .csr import ImmutableCsrMatrix


class BaseCsrOntologyGraph(OntologyGraph, metaclass=abc.ABCMeta):
    CHILD_RELATIONSHIP_CODE = 1
    PARENT_RELATIONSHIP_CODE = -1

    def __init__(self, root: NODE,
                 nodes: np.ndarray,
                 adjacency_matrix: ImmutableCsrMatrix):
        self._root = validate_instance(root, TermId, 'root')
        self._nodes = validate_instance(nodes, np.ndarray, 'nodes')
        self._adjacency_matrix = validate_instance(adjacency_matrix, ImmutableCsrMatrix, 'adjacency_matrix')

    @property
    def root(self) -> NODE:
        return self._root

    def get_children(self, source: typing.Union[str, NODE, Identified],
                     include_source: bool = False) -> typing.Iterable[NODE]:
        # In other words, find a row in the CSR corresponding to the `source`
        # and retrieve the columns to which the `source` is the PARENT.
        return map(self._get_node_for_idx,
                   self._get_node_indices_with_relationship(source,
                                                            BaseCsrOntologyGraph.PARENT_RELATIONSHIP_CODE,
                                                            include_source))

    def get_descendants(self, source: typing.Union[str, NODE, Identified],
                        include_source: bool = False) -> typing.Iterable[NODE]:
        # See `self.get_children()` for explanation of `BaseCsrOntologyGraph.PARENT_RELATIONSHIP_CODE`.
        return map(self._get_node_for_idx,
                   self._traverse_graph(source,
                                        BaseCsrOntologyGraph.PARENT_RELATIONSHIP_CODE,
                                        include_source))

    def get_parents(self, source: typing.Union[str, NODE, Identified],
                    include_source: bool = False) -> typing.Iterable[NODE]:
        # In other words, find a row in the CSR corresponding to the `source`
        # and retrieve the columns to which the `source` is the CHILD.
        return map(self._get_node_for_idx,
                   self._get_node_indices_with_relationship(source,
                                                            BaseCsrOntologyGraph.CHILD_RELATIONSHIP_CODE,
                                                            include_source))

    def get_ancestors(self, source: typing.Union[str, NODE, Identified],
                      include_source: bool = False) -> typing.Iterable[NODE]:
        # See `self.get_parents()` for explanation of `BaseCsrOntologyGraph.CHILD_RELATIONSHIP_CODE`.
        return map(self._get_node_for_idx,
                   self._traverse_graph(source,
                                        BaseCsrOntologyGraph.CHILD_RELATIONSHIP_CODE,
                                        include_source))

    def is_leaf(self, node: typing.Union[str, NODE, Identified]) -> bool:
        for _ in self._get_node_indices_with_relationship(node, BaseCsrOntologyGraph.PARENT_RELATIONSHIP_CODE, False):
            return False
        return True

    def __contains__(self, item: NODE) -> bool:
        return self._get_idx_for_node(item) is not None

    def __iter__(self) -> typing.Iterator[NODE]:
        return iter(self._nodes)

    def _traverse_graph(self, source: typing.Union[str, NODE, Identified],
                        relationship,
                        include_source: bool) -> typing.Generator[int, None, None]:
        source: TermId = self._map_to_term_id(source)
        seen: set[int] = set()
        buffer: typing.Deque[int] = deque()

        # Init
        for idx in self._get_node_indices_with_relationship(source, relationship, include_source):
            seen.add(idx)
            buffer.append(idx)

        # Loop
        while buffer:
            current = buffer.popleft()
            for idx in self._get_cols_with_relationship(current, relationship):
                if idx not in seen:
                    seen.add(idx)
                    buffer.append(idx)

            yield current

    def _get_node_indices_with_relationship(self, source: typing.Union[str, NODE, Identified],
                                            relationship,
                                            include_source: bool) -> typing.Generator[int, None, None]:
        source: TermId = self._map_to_term_id(source)
        row_idx = self._get_idx_for_node(source)
        if row_idx is None:
            raise ValueError(f'Term ID not found in the graph: {source.value}')

        if include_source:
            yield row_idx

        for idx in self._get_cols_with_relationship(row_idx, relationship):
            yield idx

    def _get_cols_with_relationship(self, idx: typing.Optional[int],
                                    relationship) -> typing.Generator[int, None, None]:
        if idx is None:
            return
        col_indices = self._adjacency_matrix.col_indices_of_val(idx, relationship)
        for col_idx in col_indices:
            yield int(col_idx)  # Numpy returns np.i64 but we need int

    @abc.abstractmethod
    def _get_idx_for_node(self, node: NODE) -> typing.Optional[int]:
        pass

    def _get_node_for_idx(self, idx: int) -> NODE:
        return self._nodes[idx]

    def __repr__(self):
        return f"BaseCsrOntologyGraph(root={self._root}, n_nodes={len(self._nodes)})"


class SimpleCsrOntologyGraph(BaseCsrOntologyGraph):
    """
    An implementation of :class:`OntologyGraph` that uses a :class:`dict` to map a NODE to adjacency matrix index.
    """

    def __init__(self, root: NODE,
                 nodes: np.ndarray,
                 adjacency_matrix: ImmutableCsrMatrix):
        super().__init__(root, nodes, adjacency_matrix)
        self._node_to_idx = {node: idx for idx, node in enumerate(nodes)}
        warnings.warn('SimpleCsrOntologyGraph will be removed from the public API in v1.0.0',
                      DeprecationWarning, stacklevel=2)

    def _get_idx_for_node(self, node: NODE) -> typing.Optional[int]:
        return self._node_to_idx[node]


class BisectPoweredCsrOntologyGraph(BaseCsrOntologyGraph):
    """
    An implementation of :class:`OntologyGraph` that uses binary search to look up NODE index.
    """

    def __init__(self, root: NODE,
                 nodes: np.ndarray,
                 adjacency_matrix: ImmutableCsrMatrix,
                 skip_validation: bool = False):
        super().__init__(root, nodes, adjacency_matrix)
        if not skip_validation:
            # NOTE - the `nodes` array ABSOLUTELY MUST be a sorted.
            # Otherwise, the traversal functionality will not work!
            check_items_are_sorted(nodes)

    def _get_idx_for_node(self, node: NODE) -> typing.Optional[int]:
        idx = bisect.bisect_left(self._nodes, node)
        if idx != len(self._nodes) and self._nodes[idx] == node:
            return idx
        return None


def check_items_are_sorted(items: typing.Iterable[NODE]):
    """
    Check that items are sorted in ascending order according to their natural ordering.
    :param items: an iterable of items.
    :return: `None` if the items are sorted or raise :class:`ValueError` if the items is unsorted.
    """
    previous = None
    for i, node in enumerate(items):
        if previous and node < previous:
            raise ValueError(f'Unsorted sequence. Item #{i} ({node}) was less than #{i - 1} ({previous})')
        previous = node



import abc
import enum
import typing

from hpotk.model import Identified, Named, TermId
from hpotk.model import CURIE_OR_TERM_ID
from ._api import FrequencyAwareFeature, AnnotatedItem, AnnotatedItemContainer


class EvidenceCode(enum.Enum):
    """
    An enumeration with evidence codes.
    """

    IEA = enum.auto()
    """
    Inferred from electronic evidence.
    """

    TAS = enum.auto()
    """
    Traceable author statement.
    """

    PCS = enum.auto()
    """
    Published clinical study.
    """

    @staticmethod
    def parse(value: str):
        """
        Parse evidence code from `str` value.

        :param value: a str with the evidence code.
        :return: the parsed enum member or `None` if `value` is not valid `EvidenceCode` value.
        """
        value = value.upper()
        if value == 'IEA':
            return EvidenceCode.IEA
        elif value == 'TAS':
            return EvidenceCode.TAS
        elif value == 'PCS':
            return EvidenceCode.PCS
        else:
            return None


class Sex(enum.Enum):
    """
    An enum representing values of apparent biological sex.

    .. note::

      We do not attempt to model all contemporary complexities of the sex.
    """

    UNKNOWN = enum.auto()
    MALE = enum.auto()
    FEMALE = enum.auto()

    @staticmethod
    def parse(value: str):
        """
        Parse :class:`Sex` from a `str` value.

        :param value: a `str` with the sex code.
        :return: the parsed enum member or `None` if `value` is not valid :class:`Sex` value.
        """
        value = value.upper()
        if value == 'MALE':
            return Sex.MALE
        elif value == 'FEMALE':
            return Sex.FEMALE
        elif value == 'UNKNOWN':
            return Sex.UNKNOWN
        else:
            return None


class AnnotationReference(Identified):

    def __init__(self, identifier: TermId,
                 evidence_code: EvidenceCode):
        if not isinstance(identifier, TermId):
            raise ValueError(f'Identifier {identifier} is not a `TermId`')
        if not isinstance(evidence_code, EvidenceCode):
            raise ValueError(f'Evidence code {evidence_code} is not an `EvidenceCode`')
        self._identifier = identifier
        self._evidence_code = evidence_code

    @property
    def identifier(self) -> TermId:
        return self._identifier

    @property
    def evidence_code(self) -> EvidenceCode:
        return self._evidence_code

    def __eq__(self, other):
        return isinstance(other, AnnotationReference) \
            and self.identifier == other.identifier \
            and self.evidence_code == other.evidence_code

    def __hash__(self):
        return hash((self._identifier, self._evidence_code))

    def __str__(self):
        return f"AnnotationReference(" \
               f"identifier={self._identifier}, " \
               f"evidence_code={self._evidence_code})"

    def __repr__(self):
        return f"AnnotationReference(" \
               f"identifier={repr(self._identifier)}, " \
               f"evidence_code={repr(self._evidence_code)})"


class HpoDiseaseAnnotation(Identified, FrequencyAwareFeature, metaclass=abc.ABCMeta):
    """
    `HpoDiseaseAnnotation` models data of a single disease annotation.

    The annotation has the following attributes:

    * `identifier` - annotation ID, e.g. `HP:0001250`
    * frequency-related attributes of the annotation, such as `frequency` (see :class:`FrequencyAwareFeature` for more info)
    * `references` - a sequence of cross-references that support presence/absence of the annotation
    * `modifiers` - a sequence of clinical modifiers of the annotation, such as age of onset, severity, laterality, ...
    """

    @property
    @abc.abstractmethod
    def references(self) -> typing.Sequence[AnnotationReference]:
        """
        :return: a list of annotation references that support presence/absence of the disease annotation
        """
        pass

    @property
    @abc.abstractmethod
    def modifiers(self) -> typing.Sequence[TermId]:
        """
        :return: a list of disease annotation modifiers
        """
        pass

    def __str__(self):
        return f"HpoDiseaseAnnotation(" \
               f"identifier={self.identifier.value}, " \
               f"frequency={self.numerator}/{self.denominator}, " \
               f"references={self.references}, " \
               f"modifiers={self.modifiers})"

    def __repr__(self):
        return f"HpoDiseaseAnnotation(" \
               f"identifier={self.identifier}, " \
               f"numerator={self.numerator}, " \
               f"denominator={self.denominator}, " \
               f"references={self.references}, " \
               f"modifiers={self.modifiers})"


class HpoDisease(AnnotatedItem[HpoDiseaseAnnotation], Identified, Named, metaclass=abc.ABCMeta):
    """
    `HpoDisease` represents a computational model of a rare disease.

    The model includes attributes:

    * `identifier` - disease ID, e.g. `OMIM:256000`
    * `name` - human-readable disease name, e.g. `LEIGH SYNDROME; LS`
    * `annotations` - the phenotype annotations of the disease. See :class:`AnnotatedItem` for more details on
      *all* annotation-related methods
    * `modes_of_inheritance` - a collection of the modes of inheritance associated with the disease

    """

    @property
    @abc.abstractmethod
    def modes_of_inheritance(self) -> typing.Collection[TermId]:
        """
        :return: a collection of modes of inheritance associated with the disease.
        """
        pass

    def __str__(self):
        return f"HpoDisease(" \
               f"identifier={self.identifier}, " \
               f"name={self.name}, " \
               f"n_annotations={len(self.annotations)})"


class HpoDiseases(AnnotatedItemContainer[HpoDiseaseAnnotation], metaclass=abc.ABCMeta):
    """
    A container for HPO diseases that allows iteration over all diseases,
    knows about the number of diseases in the container, and supports retrieval of the disease by its identifier.
    """

    @abc.abstractmethod
    def __getitem__(self, disease_id: CURIE_OR_TERM_ID) -> typing.Optional[HpoDisease]:
        """
        Get :class:`HpoDisease` based on given `disease_id` or `None` if the disease for the identifier is not present
        in the container.

        :param disease_id: a CURIE `str` or a :class:`TermId` of the disease identifier.
        :return: :class:`HpoDisease` or `None`
        """
        pass

    def __str__(self):
        return f"HpoDiseases(n_diseases={len(self)}, " \
               f"version={self.version})"



import re
import json
import typing
import logging

from hpotk.model import TermId, MinimalTerm, Term
from hpotk.graph import OntologyGraph
from hpotk.graph import GraphFactory, CsrIndexedGraphFactory, OWL_THING
from hpotk.ontology import MinimalOntology, Ontology, create_ontology, create_minimal_ontology
from hpotk.util import open_text_io_handle_for_reading

from ._model import create_node, create_edge
from ._model import Node, Edge, NodeType
from ._factory import MinimalTermFactory, TermFactory, ObographsTermFactory, MINIMAL_TERM

logger = logging.getLogger(__name__)

# TODO - verify PURL works for other ontologies than HPO
# A pattern to match an obolibrary PURL. The PURL should is expected to have 3 parts: `prefix`, `id`, and `curie`
# The `curie` is `prefix` + '_' + `id`.
PURL_PATTERN = re.compile(r'http://purl\.obolibrary\.org/obo/(?P<curie>(?P<prefix>\w+)_(?P<id>\d{7}))')
DATE_PATTERN = re.compile(r'.*/(?P<date>\d{4}-\d{2}-\d{2})/.*')


def load_minimal_ontology(file: typing.Union[typing.IO, str],
                          term_factory: ObographsTermFactory[MinimalTerm] = MinimalTermFactory(),
                          graph_factory: GraphFactory = CsrIndexedGraphFactory()) -> MinimalOntology:
    return _load_impl(file, term_factory, graph_factory, create_minimal_ontology)


def load_ontology(file: typing.Union[typing.IO, str],
                  term_factory: ObographsTermFactory[Term] = TermFactory(),
                  graph_factory: GraphFactory = CsrIndexedGraphFactory()) -> Ontology:
    return _load_impl(file, term_factory, graph_factory, create_ontology)


def _load_impl(file: typing.Union[typing.IO, str],
               term_factory: ObographsTermFactory[MinimalTerm],
               graph_factory: GraphFactory,
               ontology_creator):
    hpo = get_hpo_graph(file)
    logger.debug("Extracting ontology terms")
    id_to_term_id, terms = extract_terms(hpo['nodes'], term_factory)
    logger.debug("Creating the edge list")
    edge_list = create_edge_list(hpo['edges'], id_to_term_id)
    logger.debug("Building ontology graph")
    graph: OntologyGraph = graph_factory.create_graph(edge_list)
    if graph.root == OWL_THING:
        # TODO - consider adding Owl thing into terms list
        pass
    version = extract_ontology_version(hpo['meta'])
    logger.debug("Assembling the ontology")
    ontology = ontology_creator(graph, terms, version)
    logger.debug("Done")
    return ontology


def get_hpo_graph(file: typing.Union[typing.IO, str]):
    with open_text_io_handle_for_reading(file) as fh:
        document = json.load(fh)
    if not isinstance(document, dict):
        raise ValueError(f'The JSON document should have been a dict but was {type(document)}')
    if 'graphs' not in document:
        raise ValueError(f'Did not find the `graphs` attribute in the JSON document')
    graphs = document['graphs']
    if not isinstance(graphs, typing.Sequence):
        raise ValueError(f'`graphs` JSON attribute is not a sequence')
    if len(graphs) < 1:
        raise ValueError(f'`graphs` JSON attribute is empty')
    elif len(graphs) == 1:
        # The happy path
        return graphs[0]
    else:
        raise ValueError(f'We expect exactly 1 graph but there are {len(graphs)} graphs in the JSON document')


def extract_terms(nodes: typing.Iterable[dict],
                  term_factory: ObographsTermFactory[MINIMAL_TERM]) \
        -> typing.Tuple[typing.Mapping[str, TermId], typing.Sequence[MINIMAL_TERM]]:
    curie_to_term: typing.Dict[str, TermId] = {}
    terms: typing.List[Term] = []
    for data in nodes:
        # 1) map data to `Node`
        node: Node = create_node(data)

        # 2) we only work with class Nodes
        if not node or node.type != NodeType.CLASS:
            continue

        # 3) check if PURL is OK
        curie = extract_curie_from_purl(node.id)
        if not curie:
            logger.debug(f'Unable to parse PURL {node.id} into CURIE')
            continue
        term_id = TermId.from_curie(curie)
        if term_id.prefix != 'HP':
            logger.debug(f'Skipping non-HPO term {term_id.value}')
            continue

        curie_to_term[curie] = term_id

        # 4) create the `Term`
        term = term_factory.create_term(term_id, node)
        if term:
            terms.append(term)

    return curie_to_term, terms


def create_edge_list(edges: typing.Iterable[typing.Dict[str, str]],
                     curie_to_termid: typing.Mapping[str, TermId]) -> typing.List[typing.Tuple[TermId, TermId]]:
    edge_list: typing.List[typing.Tuple[TermId, TermId]] = []
    for data in edges:
        edge: Edge = create_edge(data)

        # We only care about `is_a` relationships.
        if edge.pred != 'is_a':
            logger.debug(f'Skipping edge with pred {edge.pred}!=\'is_a\'')
            continue

        # Get source and destination.
        try:
            curie = extract_curie_from_purl(edge.sub)
            src: TermId = curie_to_termid[curie]
        except KeyError:
            logger.warning(f'Source edge {edge.sub} was not found in terms')
            # TODO - maybe we should even abort?
            continue

        try:
            curie: str = extract_curie_from_purl(edge.obj)
            dest: TermId = curie_to_termid[curie]
        except KeyError:
            logger.warning(f'Destination edge {edge.obj} was not found in terms')
            # TODO - maybe we should even abort?
            continue

        edge_list.append((src, dest))

    return edge_list


def extract_curie_from_purl(purl: str) -> typing.Optional[str]:
    """
    Parse HPO PURL (e.g. `http://purl.obolibrary.org/obo/HP_0002813`) into the CURIE (e.g. `HP_0002813`).

    Returns the CURIE `str` or `None` if the PURL is mis-formatted.
    """
    matcher = PURL_PATTERN.match(purl)
    return matcher.group('curie') if matcher else None


def extract_ontology_version(meta: dict) -> typing.Optional[str]:
    if 'version' in meta:
        # A line like this:
        # 'http://purl.obolibrary.org/obo/hp/releases/2022-10-05/hp.json'
        match = DATE_PATTERN.search(meta['version'])
        if match:
            return match.group('date')
        else:
            logger.debug(f'Could not find a date pattern in version {meta["version"]}')
            return None
    elif 'basicPropertyValues' in meta:
        for bpv in meta['basicPropertyValues']:
            if 'pred' in bpv and 'val' in bpv:
                if bpv['pred'].endswith('#versionInfo'):
                    # An item like this:
                    # {
                    #   "pred": "http://www.w3.org/2002/07/owl#versionInfo",
                    #   "val": "2022-10-05"
                    # }
                    return bpv['val']

        logger.debug(f'Could not find basic property value with the version info')
        return None
    else:
        logger.debug(f'Could not determine the ontology version')
        return None



import typing

import numpy as np
import pytest

import hpotk
from ._csr_graph import OntologyGraph
from ._csr_idx_graph import IndexedOntologyGraph
from ._factory import IncrementalCsrGraphFactory, CsrIndexedGraphFactory
from ._factory import make_row_col_data


@pytest.fixture
def nodes() -> typing.Sequence[hpotk.TermId]:
    """
    Nodes for testing.
    """
    # Nodes are sorted.
    return (
        hpotk.TermId.from_curie('HP:01'),
        hpotk.TermId.from_curie('HP:010'),
        hpotk.TermId.from_curie('HP:011'),
        hpotk.TermId.from_curie('HP:0110'),
        hpotk.TermId.from_curie('HP:02'),
        hpotk.TermId.from_curie('HP:020'),
        hpotk.TermId.from_curie('HP:021'),
        hpotk.TermId.from_curie('HP:022'),
        hpotk.TermId.from_curie('HP:03'),
        hpotk.TermId.from_curie('HP:1'),
    )


@pytest.fixture
def edges() -> typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]:
    # Edges for a graph with a single root node: HP:1
    return (
        # source -> destination
        (hpotk.TermId.from_curie('HP:01'), hpotk.TermId.from_curie('HP:1')),
        (hpotk.TermId.from_curie('HP:010'), hpotk.TermId.from_curie('HP:01')),
        (hpotk.TermId.from_curie('HP:011'), hpotk.TermId.from_curie('HP:01')),
        # This one is multi-parent.
        (hpotk.TermId.from_curie('HP:0110'), hpotk.TermId.from_curie('HP:010')),
        (hpotk.TermId.from_curie('HP:0110'), hpotk.TermId.from_curie('HP:011')),

        (hpotk.TermId.from_curie('HP:02'), hpotk.TermId.from_curie('HP:1')),
        (hpotk.TermId.from_curie('HP:020'), hpotk.TermId.from_curie('HP:02')),
        (hpotk.TermId.from_curie('HP:021'), hpotk.TermId.from_curie('HP:02')),
        (hpotk.TermId.from_curie('HP:022'), hpotk.TermId.from_curie('HP:02')),

        (hpotk.TermId.from_curie('HP:03'), hpotk.TermId.from_curie('HP:1')),
    )


class TestFunctions:
    """
    Test the functions of the `_factory.py` file.
    """

    def test_make_row_col_data(self, nodes: typing.Sequence[hpotk.TermId],
                               edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]):
        """
        We must get these `row`, `col`, and `data` values from given `nodes` and `edges`.
        """
        row, col, data = make_row_col_data(nodes, edges)

        assert row == [0, 3, 5, 7, 9, 13, 14, 15, 16, 17, 20]
        assert col == [1, 2, 9, 0, 3, 0, 3, 1, 2, 5, 6, 7, 9, 4, 4, 4, 9, 0, 4, 8]
        assert data == [-1, -1, 1, 1, -1, 1, -1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, -1, -1, -1]

    @pytest.mark.skip
    def test_compare_with_scipy(self, nodes: typing.Sequence[hpotk.TermId],
                                edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]):
        """
        Compare that the row, col, data arrays are the same as those obtained from scipy.dense.csr_matrix when
        starting from a dense matrix.
        """
        # ############################## Expected values ##############################
        try:
            from scipy.sparse import csr_matrix
        except ImportError:
            raise Exception(f'We cannot run this test if scipy is not installed.')
        # This is the dense array that we expect for the nodes and edges created in `make_nodes_and_edges()`
        expected = np.array([[0, -1, -1, 0, 0, 0, 0, 0, 0, 1],
                             [1, 0, 0, -1, 0, 0, 0, 0, 0, 0],
                             [1, 0, 0, -1, 0, 0, 0, 0, 0, 0],
                             [0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
                             [0, 0, 0, 0, 0, -1, -1, -1, 0, 1],
                             [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                             [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                             [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                             [0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                             [1, 0, 0, 0, -1, 0, 0, 0, -1, 0]])
        csr = csr_matrix(expected)

        # ############################## Actual values ##############################
        row, col, data = make_row_col_data(nodes, edges)

        np.array_equal(csr.indptr, row)
        np.array_equal(csr.indices, col)
        np.array_equal(csr.data, data)


class TestIncrementalCsrGraphFactory:

    @pytest.fixture
    def factory(self) -> IncrementalCsrGraphFactory:
        return IncrementalCsrGraphFactory()

    def test_create_graph(self, factory: IncrementalCsrGraphFactory,
                          nodes: typing.Sequence[hpotk.TermId],
                          edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]
                          ):
        graph = factory.create_graph(edges)
        assert graph.root == hpotk.TermId.from_curie('HP:1')


class TestTraversalOfIncrementalCsrGraphFactory:

    @pytest.fixture
    def factory(self) -> IncrementalCsrGraphFactory:
        return IncrementalCsrGraphFactory()

    @pytest.fixture
    def graph(self, factory: IncrementalCsrGraphFactory,
              edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]) -> OntologyGraph:
        return factory.create_graph(edges)

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {"HP:01", "HP:02", "HP:03"}),
            ("HP:01", {"HP:010", "HP:011"}),
            ("HP:010", {"HP:0110"}),
            ("HP:011", {"HP:0110"}),
            ("HP:0110", {}),

            ("HP:02", {"HP:020", "HP:021", "HP:022"}),

            ("HP:03", {})
        ]
    )
    def test_get_children(self, graph: OntologyGraph,
                          source: str, expected: typing.Set[str]):
        children = set(graph.get_children(hpotk.TermId.from_curie(source)))
        assert children == set((hpotk.TermId.from_curie(curie) for curie in expected))

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {}),
            ("HP:01", {"HP:1"}),
            ("HP:010", {"HP:01"}),
            ("HP:0110", {"HP:010", "HP:011"}),

            ("HP:03", {"HP:1"})
        ]
    )
    def test_get_parents(self, graph: OntologyGraph,
                         source: str, expected: typing.Set[str]):
        children = set(graph.get_parents(hpotk.TermId.from_curie(source)))
        assert children == set((hpotk.TermId.from_curie(curie) for curie in expected))


class TestCsrIndexedGraphFactory:

    @pytest.fixture
    def factory(self) -> CsrIndexedGraphFactory:
        return CsrIndexedGraphFactory()

    def test_create_graph(self, factory: CsrIndexedGraphFactory,
                          edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]):
        graph: IndexedOntologyGraph = factory.create_graph(edges)

        assert isinstance(graph, IndexedOntologyGraph)
        # We test the functionality elsewhere



from hpotk.model import TermId

#  Descendents of Onset `HP:0003674`. The `#`s model the hierarchy.
ONSET: TermId = TermId.from_curie("HP:0003674")

#  #
ANTENATAL_ONSET: TermId = TermId.from_curie("HP:0030674")
#  ##
EMBRYONAL_ONSET: TermId = TermId.from_curie("HP:0011460")
#  ##
FETAL_ONSET: TermId = TermId.from_curie("HP:0011461")
#  ###
LATE_FIRST_TRIMESTER_ONSET: TermId = TermId.from_curie("HP:0034199")
SECOND_TRIMESTER_ONSET: TermId = TermId.from_curie("HP:0034198")
THIRD_TRIMESTER_ONSET: TermId = TermId.from_curie("HP:0034197")

#  #
CONGENITAL_ONSET: TermId = TermId.from_curie("HP:0003577")
NEONATAL_ONSET: TermId = TermId.from_curie("HP:0003623")
PUERPURAL_ONSET: TermId = TermId.from_curie("HP:4000040")

#  #
PEDIATRIC_ONSET: TermId = TermId.from_curie("HP:0410280")
#  ##
INFANTILE_ONSET: TermId = TermId.from_curie("HP:0003593")
CHILDHOOD_ONSET: TermId = TermId.from_curie("HP:0011463")
JUVENILE_ONSET: TermId = TermId.from_curie("HP:0003621")

#  #
ADULT_ONSET: TermId = TermId.from_curie("HP:0003581")
#  ##
YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0011462")
#  ###
EARLY_YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0025708")
INTERMEDIATE_YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0025709")
LATE_YOUNG_ADULT_ONSET: TermId = TermId.from_curie("HP:0025710")
#  ##
MIDDLE_AGE_ONSET: TermId = TermId.from_curie("HP:0003596")
LATE_ONSET: TermId = TermId.from_curie("HP:0003584")



import abc
import typing
import warnings

from hpotk.model import Identified, FrequencyAwareFeature, Versioned
from hpotk.model import TermId


# #####################################################################################################################
# The file describes generic schema for annotations, annotated items, and annotated item containers.
# An example application includes phenotypic features (annotations), diseases or samples (annotated items),
# and OMIM corpus (annotated item containers).
# #####################################################################################################################


class AnnotationBase(Identified, FrequencyAwareFeature, metaclass=abc.ABCMeta):
    pass


ANNOTATION = typing.TypeVar('ANNOTATION', bound=AnnotationBase)
"""
A world item annotation with an identifier and present or excluded state.
"""


class AnnotatedItem(typing.Generic[ANNOTATION], metaclass=abc.ABCMeta):
    """
    An item that can be annotated with ontology terms. For instance, a disease can be annotated with phenotypic features,
    items from HPO ontology.
    """

    @property
    @abc.abstractmethod
    def annotations(self) -> typing.Collection[ANNOTATION]:
        """
        :return: a collection of :class:`ANNOTATION` objects for the annotated item.
        """
        pass

    def present_annotations(self) -> typing.Iterable[ANNOTATION]:
        """
        :return: an iterable over *present* annotations.
        """
        return filter(lambda a: a.is_present, self.annotations)

    def absent_annotations(self) -> typing.Iterable[ANNOTATION]:
        """
        :return: an iterable over *absent* annotations.
        """
        return filter(lambda a: a.is_absent, self.annotations)


ANNOTATED_ITEM = typing.TypeVar('ANNOTATED_ITEM', bound=AnnotatedItem)
"""
World item that is annotated with an :class:`ANNOTATION`.
"""


class AnnotatedItemContainer(typing.Generic[ANNOTATED_ITEM],
                             typing.Iterable[ANNOTATED_ITEM],
                             typing.Sized,
                             Versioned,
                             metaclass=abc.ABCMeta):
    """
    Container for items that can be annotated with ontology terms.

    For instance, if OMIM disease is an item type and phenotypic feature is the annotation type,
    then a corpus of OMIM diseases corresponds to a container of annotated items.
    """

    @property
    def items(self) -> typing.Collection[ANNOTATED_ITEM]:
        """
        :return: an iterable over container items.
        """
        # REMOVE(v1.0.0)
        warnings.warn(f'`items` property has been deprecated and will be removed in v1.0.0. '
                      f'Iterate directly over the container.',
                      DeprecationWarning, stacklevel=2)
        return list(self)

    def item_ids(self) -> typing.Iterable[TermId]:
        """
        :return: an iterable over all item identifiers.
        """
        return map(lambda item: item.identifier, self)




"""
The `hpotk.util.store` package provides
"""

from ._api import OntologyType, OntologyStore
from ._config import configure_ontology_store

__all__ = [
    'OntologyType', 'OntologyStore',
    'configure_ontology_store',
]



import abc


class TermId(metaclass=abc.ABCMeta):
    """
    `TermId` is an identifier of an ontology concept.

    `TermId` consists of a *prefix* and *id* that are separated by a delimiter:

    .. doctest:: term-id

      >>> term_id = TermId.from_curie('HP:0001250')
      >>> assert term_id.prefix == 'HP'
      >>> assert term_id.id == '0001250'

    The `TermId` has a natural ordering which compares two IDs first based on prefix and then value.
    Both comparisons are lexicographic.
    """

    @staticmethod
    def from_curie(curie: str):
        """
        Create a `TermId` from a compact URI (CURIE).

        The prefix and id of a `TermId` must be separated either by a colon ``:`` or an underscore ``_``.

        .. doctest:: term-id

          >>> term_id = TermId.from_curie('HP:0001250')
          >>> term_id.value
          'HP:0001250'

        The parsing will forget the original delimiter. The `value` always joins the *prefix* and *id* with ``:``.

        .. doctest:: term-id

          >>> ncit = TermId.from_curie('NCIT_C3117')
          >>> ncit.value
          'NCIT:C3117'

        The ``:`` has higher priority than ``_``, and it will be used as delimiter.

        .. doctest:: term-id

          >>> snomed = TermId.from_curie('SNOMEDCT_US:128613002')
          >>> snomed.prefix
          'SNOMEDCT_US'
          >>> snomed.id
          '128613002'

        :param curie: a CURIE `str` to be parsed.
        :return: the created `TermId`.
        :raises: `ValueError` if the value is mis-formatted.
        """
        if curie is None:
            raise ValueError(f'Curie must not be None')
        try:
            idx = curie.index(':')
        except ValueError:
            try:
                idx = curie.index('_')
            except ValueError:
                raise ValueError(f'The CURIE {curie} has no colon `:` or underscore `_`')

        return DefaultTermId(idx=idx, value=curie)

    @property
    @abc.abstractmethod
    def prefix(self) -> str:
        """
        Get `prefix` of the ontology concept.

        .. doctest:: term-id

          >>> term_id = TermId.from_curie('HP:1234567')
          >>> term_id.prefix
          'HP'
        """
        pass

    @property
    @abc.abstractmethod
    def id(self) -> str:
        """
        Get `id` of the ontology concept.

        .. doctest:: term-id

          >>> term_id = TermId.from_curie('HP:1234567')
          >>> term_id.id
          '1234567'
        """
        pass

    @property
    def value(self) -> str:
        """
        Get concept value consisting of `self.prefix` and `self.value`.

        .. doctest:: term-id

          >>> term_id = TermId.from_curie('HP:1234567')
          >>> term_id.value
          'HP:1234567'
        """
        return self.prefix + ':' + self.id

    @staticmethod
    def _calculate_hash(prefix: str, id: str) -> int:
        """
        Calculate hash of a term ID from the members.

        :param prefix: the prefix part of the term ID (e.g. `HP` for `HP:1234567`) as a `str.
        :param id: the id part of the term ID (e.g. `1234567` for `HP:1234567`) as a str.
        :return: a hash as an `int`.
        """
        return hash((prefix, id))

    def __hash__(self) -> int:
        return self._calculate_hash(self.prefix, self.id)

    def __eq__(self, other):
        return isinstance(other, TermId) \
            and self.prefix == other.prefix \
            and self.id == other.id

    def __lt__(self, other):
        if isinstance(other, TermId):
            if self.prefix == other.prefix:
                return self.id < other.id
            else:
                return self.prefix < other.prefix
        else:
            return NotImplemented

    def __str__(self):
        return self.value


class DefaultTermId(TermId):
    """
    A default implementation of :class:`TermId` that stores the index of the delimiter and the value as a string
    and caches the hash value.
    """

    def __init__(self, value: str, idx: int):
        self._value = value
        self._idx = idx
        self._hash = self._calculate_hash(prefix=value[:idx], id=value[idx + 1:])

    @property
    def prefix(self) -> str:
        return self._value[:self._idx]

    @property
    def id(self) -> str:
        return self._value[self._idx + 1:]

    def __repr__(self):
        return f'DefaultTermId(idx={self._idx}, value={self._value})'

    def __hash__(self) -> int:
        return self._hash

    # TODO - make specific HPO TermId


class SimpleTermId(TermId):
    """
    The simplest possible implementation of a `TermId` that stores the entire curie and the position
    of the delimiter that separates the prefix and the id.
    """

    def __init__(self, value: str, idx: int):
        self._value = value
        self._idx = idx

    @property
    def prefix(self) -> str:
        return self._value[:self._idx]

    @property
    def id(self) -> str:
        return self._value[self._idx + 1:]

    def __repr__(self):
        return f'SimpleTermId(idx={self._idx}, value={self._value})'



# TODO[v1.0.0] - remove the deprecated methods
from ._traversal import get_ancestors, get_parents
from ._traversal import get_children, get_descendants
from ._traversal import exists_path
from . import similarity



import abc
import bisect
import logging
import typing
import warnings
from collections import defaultdict

import numpy as np

from hpotk.model import TermId
from ._api import OntologyGraph, IndexedOntologyGraph, NODE, OWL_THING
from ._csr_graph import SimpleCsrOntologyGraph, BisectPoweredCsrOntologyGraph
from ._csr_idx_graph import StaticCsrArray, CsrData, CsrIndexedOntologyGraph
from .csr import CsrMatrixBuilder, ImmutableCsrMatrix

# A newtype for stronger typing. We use these in `GraphFactory` below.
DirectedEdge = typing.Tuple[TermId, TermId]
GRAPH = typing.TypeVar('GRAPH', bound=OntologyGraph)


class GraphFactory(typing.Generic[GRAPH], metaclass=abc.ABCMeta):
    """
    Graph factory creates a graph from a list of `TermId` pairs.
    """

    def __init__(self):
        self._logger = logging.getLogger(__name__)

    @abc.abstractmethod
    def create_graph(self, edge_list: typing.Sequence[DirectedEdge]) -> GRAPH:
        """
        Create graph from edge list.

        :param edge_list: a sequence of directed edges - tuples where the first item is the source
          and the second item is the destination.
        :return: the graph
        """
        pass


class AbstractCsrGraphFactory(GraphFactory[OntologyGraph], metaclass=abc.ABCMeta):

    def create_graph(self, edge_list: typing.Sequence[DirectedEdge]) -> GRAPH:
        # Find root node
        self._logger.debug('Creating ontology graph from %d edges', len(edge_list))
        root, edge_list = _phenol_find_root(edge_list)
        self._logger.debug('Found root %s', root.value)

        # Prepare node list. We MUST sort the list, otherwise building of the IncrementalCsrMatrix won't work.
        nodes = get_array_of_unique_and_sorted_nodes(edge_list)
        self._logger.debug('Extracted %d nodes', len(nodes))

        # Build the adjacency matrix
        self._logger.debug('Building sparse adjacency matrix')
        cm = self._build_adjacency_matrix(nodes, edge_list)
        # Assemble the ontology
        self._logger.debug('Finalizing the ontology graph')
        return BisectPoweredCsrOntologyGraph(root, nodes, cm)

    @abc.abstractmethod
    def _build_adjacency_matrix(self, nodes: typing.Sequence[TermId],
                                edges: typing.Sequence[DirectedEdge]) -> ImmutableCsrMatrix:
        pass


class CsrGraphFactory(AbstractCsrGraphFactory):
    """
    A factory for creating `OntologyGraph` that is backed by a compressed sparse row (CSR) connectivity matrix.

    .. deprecated::
      Use :class:`hpotk.graph.IncrementalCsrGraphFactory` instead.
    """

    def __init__(self):
        # REMOVE(v1.0.0)
        warnings.warn('CsrGraphFactory was deprecated and will be removed in v1.0.0', DeprecationWarning, stacklevel=2)
        super().__init__()

    def _build_adjacency_matrix(self, nodes: typing.Sequence[TermId],
                                edges: typing.Sequence[DirectedEdge]):
        node_to_idx = {node: idx for idx, node in enumerate(nodes)}
        builder = CsrMatrixBuilder(shape=(len(nodes), len(nodes)))
        for edge in edges:
            src_idx = node_to_idx[edge[0]]
            dest_idx = node_to_idx[edge[1]]
            builder[src_idx, dest_idx] = SimpleCsrOntologyGraph.CHILD_RELATIONSHIP_CODE
            builder[dest_idx, src_idx] = SimpleCsrOntologyGraph.PARENT_RELATIONSHIP_CODE
        return ImmutableCsrMatrix(builder.row,
                                  builder.col,
                                  builder.data,
                                  builder.shape,
                                  dtype=int)


class CsrIndexedGraphFactory(GraphFactory[IndexedOntologyGraph]):
    """
    `CsrIndexedGraphFactory` builds an :class:`IndexedOntologyGraph` that is backed by two CSR arrays for storing
    the parent and child nodes of an ontology graph node.
    """

    def __init__(self):
        super().__init__()

    def create_graph(self, edge_list: typing.Sequence[DirectedEdge]) -> GRAPH:
        # Find root node
        self._logger.debug('Creating ontology graph from %d edges', len(edge_list))
        root, edge_list = _phenol_find_root(edge_list)
        self._logger.debug('Found root %s', root.value)

        # Prepare node list. We MUST sort the list, otherwise building of the IncrementalCsrMatrix won't work.
        nodes = get_array_of_unique_and_sorted_nodes(edge_list)
        self._logger.debug('Extracted %d nodes', len(nodes))

        root_idx = self._find_root_idx(root, nodes)
        csr_data = self._build_csr_data(nodes, edge_list)

        return CsrIndexedOntologyGraph(root_idx, nodes, csr_data)

    @staticmethod
    def _find_root_idx(root: NODE,
                       nodes: typing.Sequence[NODE]) -> int:
        # Simple linear search for now
        for i, node in enumerate(nodes):
            if node == root:
                return i

        raise ValueError(f'Did not find root {root} in the nodes')

    def _build_csr_data(self, nodes: np.ndarray,
                        edges: typing.Sequence[DirectedEdge]) -> CsrData:
        adjacent_edges = self._find_adjacent_edges(nodes, edges)

        parent_indptr, parents = [0], []
        child_indptr, children = [0], []

        for row_idx, source in enumerate(nodes):
            adjacent = adjacent_edges.get(row_idx, ())

            for edge in adjacent:
                target_is_child = source == edge[1]  # edge[1] is the object
                target = edge[0] if target_is_child else edge[1]
                idx = _index_of_using_binary_search(nodes, target)
                if target_is_child:
                    children.append(idx)
                else:
                    parents.append(idx)

            parent_indptr.append(len(parents))
            child_indptr.append(len(children))

        parents = StaticCsrArray(parent_indptr, parents)
        children = StaticCsrArray(child_indptr, children)

        return CsrData(children=children, parents=parents)

    @staticmethod
    def _find_adjacent_edges(nodes: np.ndarray,
                             edges: typing.Sequence[DirectedEdge]) -> typing.Mapping[int, typing.Sequence[DirectedEdge]]:
        data = defaultdict(list)

        last_sub = None
        last_sub_idx = -1

        for edge in edges:
            sub = edge[0]
            if sub == last_sub:
                sub_idx = last_sub_idx
            else:
                last_sub = sub
                sub_idx = _index_of_using_binary_search(nodes, sub)
                last_sub_idx = sub_idx

            obj = edge[1]
            obj_idx = _index_of_using_binary_search(nodes, obj)

            data[sub_idx].append(edge)
            data[obj_idx].append(edge)

        return data


def get_array_of_unique_and_sorted_nodes(edge_list: typing.Sequence[DirectedEdge]) -> np.ndarray:
    edges = np.array(edge_list)
    return np.unique(edges)


def get_list_of_unique_nodes(edge_list: typing.Sequence[DirectedEdge]):
    return list(get_unique_nodes(edge_list))


def get_unique_nodes(edge_list: typing.Sequence[DirectedEdge]) -> typing.Collection[TermId]:
    nodes: typing.Set[TermId] = set()
    for edge in edge_list:
        nodes.add(edge[0])
        nodes.add(edge[1])
    return nodes


def _phenol_find_root(edge_list: typing.Sequence[DirectedEdge]) -> typing.Tuple[NODE, typing.Sequence[DirectedEdge]]:
    """
    Find an ontology root candidate - the term that is parent of all elements using `DirectedEdge` that represents
    `src` -> `is_a` -> `dst` relationship.

    The method finds term IDs with no parents. There are 3 situations that can happen:
    - no term ID with no parents are found - we throw an error
    - one term ID with no parents is found, this is the root candidate
    - two or more term IDs with no parents are found, we add `OWL_THING` as the root term.
    :param edge_list:
    :return: a tuple with the root and possibly updated edge_list
    """
    root_candidate_set: typing.Set[TermId] = set()
    remove_mark_set: typing.Set[TermId] = set()
    for edge in edge_list:
        root_candidate_set.add(edge[1])
        remove_mark_set.add(edge[0])

    candidates = root_candidate_set.difference(remove_mark_set)
    if len(candidates) == 0:
        raise ValueError(f'No root candidate found')
    if len(candidates) == 1:
        return candidates.pop(), edge_list
    else:
        # No single root candidate, so create a new one and add it into the nodes and edges
        # As per suggestion https://github.com/monarch-initiative/phenol/issues/163#issuecomment-452880405
        # We'll use owl:Thing instead of ID:0000000 so as not to potentially conflict with an existing term id.
        # TODO - we need tests!
        edge_list_copy = [val for val in edge_list]
        for candidate in candidates:
            edge = (candidate, OWL_THING)
            edge_list_copy.append(edge)

        return OWL_THING, edge_list_copy


class IncrementalCsrGraphFactory(AbstractCsrGraphFactory):
    """
    The CSR graph factory that builds the `row`, `col` and `data` in an incremental fashion.
    """

    def _build_adjacency_matrix(self, nodes: typing.Sequence[TermId],
                                edges: typing.Sequence[DirectedEdge]):
        row, col, data = make_row_col_data(nodes, edges)
        shape = (len(nodes), len(nodes))
        return ImmutableCsrMatrix(row, col, data, shape, dtype=int)


def make_row_col_data(nodes: typing.Sequence[TermId],
                      edge_list: typing.Sequence[DirectedEdge]):
    row = [0]
    col = []
    data = []
    partitioned_edges = _partition_edges(nodes, edge_list)

    for row_idx, node in enumerate(nodes):
        relevant_edges = partitioned_edges[row_idx]

        for target, relationship_code in sorted(_preprocess_edges(node, relevant_edges),
                                                key=lambda e: e[0]):  # We sort by the TermId
            idx = _index_of_using_binary_search(nodes, target)
            col.append(idx)
            data.append(relationship_code)

        row.append(len(data))

    return row, col, data


def _partition_edges(nodes: typing.Sequence[TermId],
                     edge_list: typing.Sequence[DirectedEdge]) -> typing.Mapping[int, typing.Sequence[DirectedEdge]]:
    """
    Prepare a mapping from a node index to all edges that reference the node.

    :param nodes: a sequence of nodes.
    :param edge_list: a sequence of edges.
    :return: a mapping from a node index to a sequence of edges that reference the node.
    """
    data = defaultdict(list)
    last_src = None
    last_idx = None
    for edge in edge_list:
        src = edge[0]
        if src == last_src:
            #  leverage Let's save one binary search
            src_idx = last_idx
        else:
            last_src = src
            src_idx = _index_of_using_binary_search(nodes, src)
            last_idx = src_idx
        dest = edge[1]
        dest_idx = _index_of_using_binary_search(nodes, dest)

        if src_idx is not None:
            data[src_idx].append(edge)
        if dest_idx is not None:
            data[dest_idx].append(edge)

    return data


def _preprocess_edges(source: TermId,
                      relevant_edges: typing.Iterable[DirectedEdge]):
    """
    Get a generator for yielding tuples with a term_id and a relationship code for terms/nodes that have a relationship
    with the `source`.

    :param source: a term ID for which we are generating the CSR row.
    :param relevant_edges: a sequence of edges that reference the `source` term ID
    :return: a tuple with the node/term ID and the relationship to store in the graph.
    """
    for edge in relevant_edges:
        sub, obj = edge
        if source != sub and source == obj:
            # `source` is the object. Since we are getting `is_a` relationships, the current edge must represent
            # something `is_a` `source`, hence `source` is the parent.
            yield sub, SimpleCsrOntologyGraph.PARENT_RELATIONSHIP_CODE
        elif source != obj and source == sub:
            yield obj, SimpleCsrOntologyGraph.CHILD_RELATIONSHIP_CODE
        else:
            raise ValueError(f'source {source} must either be a subject or object of the edge {edge}')


def _index_of_using_binary_search(a: typing.Sequence[TermId], x: TermId) -> typing.Optional[int]:
    idx = bisect.bisect_left(a, x)
    if idx != len(a) and a[idx] == x:
        return idx
    return None



import typing
import unittest

import ddt
import numpy as np
import pytest

import hpotk
from ._api import OntologyGraph, IndexedOntologyGraph
from ._csr_graph import BisectPoweredCsrOntologyGraph
from ._factory import CsrIndexedGraphFactory, IncrementalCsrGraphFactory
from .csr import ImmutableCsrMatrix


@pytest.fixture(scope='module')
def edges() -> typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]:
    # Edges for a graph with a single root node: HP:1
    return (
        # source -> destination
        (hpotk.TermId.from_curie('HP:01'), hpotk.TermId.from_curie('HP:1')),
        (hpotk.TermId.from_curie('HP:010'), hpotk.TermId.from_curie('HP:01')),
        (hpotk.TermId.from_curie('HP:011'), hpotk.TermId.from_curie('HP:01')),
        # This one is multi-parent.
        (hpotk.TermId.from_curie('HP:0110'), hpotk.TermId.from_curie('HP:010')),
        (hpotk.TermId.from_curie('HP:0110'), hpotk.TermId.from_curie('HP:011')),

        (hpotk.TermId.from_curie('HP:02'), hpotk.TermId.from_curie('HP:1')),
        (hpotk.TermId.from_curie('HP:020'), hpotk.TermId.from_curie('HP:02')),
        (hpotk.TermId.from_curie('HP:021'), hpotk.TermId.from_curie('HP:02')),
        (hpotk.TermId.from_curie('HP:022'), hpotk.TermId.from_curie('HP:02')),

        (hpotk.TermId.from_curie('HP:03'), hpotk.TermId.from_curie('HP:1')),
    )


@pytest.fixture(scope='module')
def indexed_graph(edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]) -> IndexedOntologyGraph:
    factory = CsrIndexedGraphFactory()
    return factory.create_graph(edges)


@pytest.fixture(scope='module')
def base_graph(edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]) -> OntologyGraph:
    factory = IncrementalCsrGraphFactory()
    return factory.create_graph(edges)


@pytest.mark.parametrize(
    'source, expected',
    [
        ("HP:1", {"HP:01", "HP:02", "HP:03"}),
        ("HP:01", {"HP:010", "HP:011"}),
        ("HP:010", {"HP:0110"}),
        ("HP:011", {"HP:0110"}),
        ("HP:0110", {}),

        ("HP:02", {"HP:020", "HP:021", "HP:022"}),

        ("HP:03", {})
    ]
)
class TestGetChildren:

    def test_get_children__indexed_graph(self,
                                         indexed_graph: IndexedOntologyGraph,
                                         source, expected):
        actual = set(indexed_graph.get_children(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_get_children__base_graph(self,
                                      base_graph: OntologyGraph,
                                      source, expected):
        actual = set(base_graph.get_children(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    'source, expected',
    [
        ('HP:1', {'HP:01', 'HP:010', 'HP:011', 'HP:0110',
                  'HP:02', 'HP:020', 'HP:021', 'HP:022',
                  'HP:03'}),
        ('HP:01', {'HP:010', 'HP:011', 'HP:0110'}),
        ('HP:010', {'HP:0110'}),
        ('HP:011', {'HP:0110'}),
        ('HP:0110', set()),

        ('HP:02', {'HP:020', 'HP:021', 'HP:022'}),
        ('HP:020', set()),
        ('HP:021', set()),
        ('HP:022', set()),
        ('HP:03', set()),
    ]
)
class TestGetDescendants:

    def test_get_descendants__indexed_graph(self,
                                            indexed_graph: IndexedOntologyGraph,
                                            source, expected):
        actual = set(indexed_graph.get_descendants(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_get_descendants__base_graph(self,
                                         base_graph: OntologyGraph,
                                         source, expected):
        actual = set(base_graph.get_descendants(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    'source, expected',
    [
        ("HP:1", {}),
        ("HP:01", {"HP:1"}),
        ("HP:010", {"HP:01"}),
        ("HP:0110", {"HP:010", "HP:011"}),

        ("HP:03", {"HP:1"})
    ]
)
class TestGetParents:

    def test_get_parents__indexed_graph(self,
                                        indexed_graph: IndexedOntologyGraph,
                                        source, expected):
        children = set(indexed_graph.get_parents(source))

        assert children == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_get_parents__base_graph(self,
                                     base_graph: OntologyGraph,
                                     source, expected):
        actual = set(base_graph.get_parents(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    'source, expected',
    [
        ('HP:1', set()),
        ('HP:01', {'HP:1'}),
        ('HP:0110', {'HP:010', 'HP:011', 'HP:01', 'HP:1'}),
        ('HP:022', {'HP:02', 'HP:1'}),
        ('HP:03', {'HP:1'}),
    ]
)
class TestGetAncestors:

    def test_indexed_graph(self,
                           indexed_graph: IndexedOntologyGraph,
                           source, expected):
        actual = set(indexed_graph.get_ancestors(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    def test_base_graph(self,
                        base_graph: OntologyGraph,
                        source, expected):
        actual = set(base_graph.get_ancestors(source))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)


@pytest.mark.parametrize(
    'sub, obj, expected',
    [
        # True examples
        ('HP:1', 'HP:01', True),

        ('HP:01', 'HP:010', True),
        ('HP:01', 'HP:011', True),
        ('HP:010', 'HP:0110', True),
        ('HP:011', 'HP:0110', True),

        ('HP:1', 'HP:02', True),
        ('HP:02', 'HP:020', True),
        ('HP:02', 'HP:021', True),
        ('HP:02', 'HP:022', True),

        ('HP:1', 'HP:03', True),

        # False examples
        ('HP:1', 'HP:1', False),
        ('HP:02', 'HP:01', False),
        ('HP:01', 'HP:020', False),
        ('HP:0110', 'HP:03', False),
    ]
)
class TestIsParentOf:

    def test_indexed_graph(self,
                           indexed_graph: IndexedOntologyGraph,
                           sub, obj, expected):
        assert indexed_graph.is_parent_of(sub, obj) is expected

    def test_base_graph(self,
                        base_graph: OntologyGraph,
                        sub, obj, expected):
        assert base_graph.is_parent_of(sub, obj) is expected


@pytest.mark.parametrize(
    'sub, obj, expected',
    [
        # True examples
        ('HP:1', 'HP:010', True),
        ('HP:1', 'HP:011', True),
        ('HP:1', 'HP:0110', True),
        ('HP:01', 'HP:0110', True),

        ('HP:1', 'HP:020', True),
        ('HP:1', 'HP:021', True),
        ('HP:1', 'HP:022', True),

        # False examples
        ('HP:1', 'HP:1', False),
        ('HP:01', 'HP:1', False),
    ]
)
class TestIsAncestorOf:

    def test_indexed_graph(self,
                           indexed_graph: IndexedOntologyGraph,
                           sub, obj, expected):
        assert indexed_graph.is_ancestor_of(sub, obj) is expected

    def test_base_graph(self,
                        base_graph: OntologyGraph,
                        sub, obj, expected):
        assert base_graph.is_ancestor_of(sub, obj) is expected


@pytest.mark.parametrize(
    'sub, obj, expected',
    [
        # True examples
        ('HP:01', 'HP:1', True),

        ('HP:010', 'HP:01', True),
        ('HP:011', 'HP:01', True),
        ('HP:0110', 'HP:010', True),
        ('HP:0110', 'HP:011', True),

        ('HP:02', 'HP:1', True),
        ('HP:020', 'HP:02', True),
        ('HP:021', 'HP:02', True),
        ('HP:022', 'HP:02', True),

        ('HP:03', 'HP:1', True),

        # False examples
        ('HP:1', 'HP:1', False),
        ('HP:01', 'HP:02', False),
        ('HP:020', 'HP:01', False),
        ('HP:03', 'HP:0110', False),
    ]
)
class TestIsChildOf:

    def test_indexed_graph(self,
                           indexed_graph: IndexedOntologyGraph,
                           sub, obj, expected):
        assert indexed_graph.is_child_of(sub, obj) is expected

    def test_base_graph(self,
                        base_graph: OntologyGraph,
                        sub, obj, expected):
        assert base_graph.is_child_of(sub, obj) is expected


@pytest.mark.parametrize(
    'sub, obj, expected',
    [
        # True examples
        ('HP:010', 'HP:1', True),
        ('HP:011', 'HP:1', True),
        ('HP:0110', 'HP:1', True),
        ('HP:0110', 'HP:01', True),

        ('HP:020', 'HP:1', True),
        ('HP:021', 'HP:1', True),
        ('HP:022', 'HP:1', True),

        # False examples
        ('HP:1', 'HP:1', False),
        ('HP:1', 'HP:01', False),
    ]
)
class TestIsDescendantOf:

    def test_indexed_graph(self,
                           indexed_graph: IndexedOntologyGraph,
                           sub, obj, expected):
        assert indexed_graph.is_descendant_of(sub, obj) is expected

    def test_base_graph(self,
                        base_graph: OntologyGraph,
                        sub, obj, expected):
        assert base_graph.is_descendant_of(sub, obj) is expected


@pytest.mark.parametrize(
    'query, expected',
    [
        ('HP:1', False),

        ('HP:01', False),
        ('HP:010', False),
        ('HP:011', False),
        ('HP:0110', True),

        ('HP:02', False),
        ('HP:020', True),
        ('HP:021', True),
        ('HP:022', True),

        ('HP:03', True),
    ]
)
class TestIsLeaf:

    def test_indexed_graph(self,
                           indexed_graph: IndexedOntologyGraph,
                           query, expected):
        assert indexed_graph.is_leaf(query) is expected

    def test_base_graph(self,
                        base_graph: OntologyGraph,
                        query, expected):
        assert base_graph.is_leaf(query) is expected


@ddt.ddt
class TestCsrOntologyGraph(unittest.TestCase):
    """
    Tests of the `OntologyGraph` API that just happen to be using `BisectPoweredCsrOntologyGraph` as the implementation.
    """

    NODES: np.ndarray
    GRAPH: BisectPoweredCsrOntologyGraph

    @classmethod
    def setUpClass(cls) -> None:
        root = hpotk.TermId.from_curie('HP:1')
        curies = [
            'HP:01', 'HP:010', 'HP:011', 'HP:0110',
            'HP:02', 'HP:020', 'HP:021', 'HP:022',
            'HP:03', 'HP:1'
        ]
        nodes = np.fromiter(map(hpotk.TermId.from_curie, curies), dtype=object)
        row = [0, 3, 5, 7, 9, 13, 14, 15, 16, 17, 20]
        col = [1, 2, 9, 0, 3, 0, 3, 1, 2, 5, 6, 7, 9, 4, 4, 4, 9, 0, 4, 8]
        data = [-1, -1, 1, 1, -1, 1, -1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, -1, -1, -1]
        am = ImmutableCsrMatrix(row, col, data, shape=(len(nodes), len(nodes)), dtype=int)

        cls.NODES = nodes
        cls.GRAPH = BisectPoweredCsrOntologyGraph(root, nodes, am)

    def test_graph_queries_work_with_identified(self):
        # tests
        self.assertTrue(self.GRAPH.is_ancestor_of(SimpleIdentified.from_curie('HP:01'),
                                                  SimpleIdentified.from_curie('HP:0110')))
        self.assertTrue(self.GRAPH.is_parent_of(SimpleIdentified.from_curie('HP:01'),
                                                SimpleIdentified.from_curie('HP:010')))
        self.assertTrue(self.GRAPH.is_child_of(SimpleIdentified.from_curie('HP:01'),
                                               SimpleIdentified.from_curie('HP:1')))
        self.assertTrue(self.GRAPH.is_descendant_of(SimpleIdentified.from_curie('HP:01'),
                                                    SimpleIdentified.from_curie('HP:1')))
        self.assertTrue(self.GRAPH.is_leaf(SimpleIdentified.from_curie('HP:03')))

        # traversal
        self.assertSetEqual(set(self.GRAPH.get_ancestors(SimpleIdentified.from_curie('HP:010'))),
                            {hpotk.TermId.from_curie('HP:01'), hpotk.TermId.from_curie('HP:1')})
        self.assertSetEqual(set(self.GRAPH.get_parents(SimpleIdentified.from_curie('HP:01'))),
                            {hpotk.TermId.from_curie('HP:1')})
        self.assertSetEqual(set(self.GRAPH.get_children(SimpleIdentified.from_curie('HP:01'))),
                            {hpotk.TermId.from_curie('HP:010'), hpotk.TermId.from_curie('HP:011')})
        self.assertSetEqual(set(self.GRAPH.get_descendants(SimpleIdentified.from_curie('HP:01'))),
                            {hpotk.TermId.from_curie('HP:010'), hpotk.TermId.from_curie('HP:011'), hpotk.TermId.from_curie('HP:0110')})

    def test_graph_queries_work_with_str(self):
        # tests
        self.assertTrue(self.GRAPH.is_ancestor_of('HP:01', 'HP:0110'))
        self.assertTrue(self.GRAPH.is_parent_of('HP:01', 'HP:010'))
        self.assertTrue(self.GRAPH.is_child_of('HP:01', 'HP:1'))
        self.assertTrue(self.GRAPH.is_descendant_of('HP:01', 'HP:1'))
        self.assertTrue(self.GRAPH.is_leaf('HP:03'))

        # traversal
        self.assertSetEqual(set(self.GRAPH.get_ancestors('HP:010')),
                            {hpotk.TermId.from_curie('HP:01'), hpotk.TermId.from_curie('HP:1')})
        self.assertSetEqual(set(self.GRAPH.get_parents('HP:01')),
                            {hpotk.TermId.from_curie('HP:1')})
        self.assertSetEqual(set(self.GRAPH.get_children('HP:01')),
                            {hpotk.TermId.from_curie('HP:010'), hpotk.TermId.from_curie('HP:011')})
        self.assertSetEqual(set(self.GRAPH.get_descendants('HP:01')),
                            {hpotk.TermId.from_curie('HP:010'), hpotk.TermId.from_curie('HP:011'), hpotk.TermId.from_curie('HP:0110')})

    @ddt.data(
        ('is_parent_of',),
        ('is_ancestor_of',),
        ('is_child_of',),
        ('is_descendant_of',),
    )
    @ddt.unpack
    def test_query_methods_with_unknown_source(self, func_name):
        existing = hpotk.TermId.from_curie('HP:1')
        unknown = hpotk.TermId.from_curie('HP:999')

        func = getattr(self.GRAPH, func_name)
        with self.assertRaises(ValueError) as ctx:
            func(existing, unknown)
        self.assertEqual('Term ID not found in the graph: HP:999', ctx.exception.args[0])

    def test_traversal_methods_produce_iterators(self):
        whatever = hpotk.TermId.from_curie('HP:1')
        self.assertIsInstance(self.GRAPH.get_parents(whatever), typing.Iterator)
        self.assertIsInstance(self.GRAPH.get_children(whatever), typing.Iterator)
        self.assertIsInstance(self.GRAPH.get_ancestors(whatever), typing.Iterator)
        self.assertIsInstance(self.GRAPH.get_descendants(whatever), typing.Iterator)


class SimpleIdentified(hpotk.model.Identified):

    @staticmethod
    def from_curie(curie: str):
        return SimpleIdentified(hpotk.TermId.from_curie(curie))

    def __init__(self, identifier: hpotk.TermId):
        self._id = identifier

    @property
    def identifier(self) -> hpotk.TermId:
        return self._id



import abc
import typing
import warnings

from ._term_id import TermId


class Identified(metaclass=abc.ABCMeta):
    """
    A mixin for entities that have an identifier in form of a :class:`TermId`.
    """

    @property
    @abc.abstractmethod
    def identifier(self) -> TermId:
        """
        Get the identifier.
        """
        pass


class ObservableFeature(metaclass=abc.ABCMeta):
    """
    `ObservableFeature` represents a feature that can be either in a *present* or an *excluded* state
    in the investigated item(s).

    The simplest case is the presence or absence of a phenotypic feature, such as hexadactyly, in a study subject.
    """

    @property
    @abc.abstractmethod
    def is_present(self) -> bool:
        """
        Test if the feature was observed in one or more items.

        :return: `True` if the feature was observed in one or more items.
        """
        pass

    @property
    def is_absent(self) -> bool:
        """
        Test if the feature is excluded.

        .. deprecated:: 0.2.1
          Use :py:func:`is_excluded` instead.

        """
        # REMOVE[v1.0.0]
        warnings.warn("`is_absent` was deprecated and will be removed in v1.0.0. Use `is_excluded` instead",
                      DeprecationWarning, stacklevel=2)
        return self.is_excluded

    @property
    def is_excluded(self) -> bool:
        """
        Test if the feature was not observed in any of the items.

        :return: `True` if the feature was observed in *none* of the annotated item(s), and was, therefore, excluded.
        """
        return not self.is_present


class FrequencyAwareFeature(ObservableFeature, metaclass=abc.ABCMeta):
    """
    `FrequencyAwareFeature` entities describe the frequency of a feature in one or more annotated items.

    This is on top of the dichotomous state of :class:`ObservableFeature`, where the feature is present or excluded.

    For instance, we can represent the feature frequency in a collection of items, such as presence of
    a phenotypic feature, such as hexadactyly, in a cohort.

    The absolute counts are accessible via `numerator` and `denominator` properties.

    **IMPORTANT**: the implementor must ensure the following invariants:

     * the `numerator` must be a non-negative `int`
     * the `denominator` must be a positive `int`

    Use the convenience static method :py:func:`check_numerator_and_denominator` to check
    the invariants.
    """

    @property
    @abc.abstractmethod
    def numerator(self) -> int:
        """
        Get the numerator, a non-negative `int` representing the count of annotated items where the annotation
        was present.
        """
        pass

    @property
    @abc.abstractmethod
    def denominator(self) -> int:
        """
        Get the denominator, a positive `int` representing the total count of annotated items investigated
        for presence/absence of an annotation.
        """
        pass

    def frequency(self) -> float:
        """
        Get a `float` in range :math:`[0, 1]` representing the ratio of the annotation in the annotated item(s).
        """
        return self.numerator / self.denominator

    @property
    def is_present(self) -> bool:
        return self.numerator != 0

    @property
    def is_excluded(self) -> bool:
        return self.numerator == 0

    @staticmethod
    def check_numerator_and_denominator(numerator: int, denominator: int) -> None:
        """
        Check if the `numerator` and `denominator` satisfy the requirements described in :class:`FrequencyAwareFeature`.

        :return: `None` if the check passes or raises a `ValueError` if the `numerator` or `denominator` contain
          invalid values.
        """
        if not isinstance(numerator, int) or numerator < 0:
            raise ValueError(f'Numerator {numerator} must be a non-negative `int`')
        if not isinstance(denominator, int) or denominator <= 0:
            raise ValueError(f'Denominator {denominator} must be a positive `int`')


class Named(metaclass=abc.ABCMeta):
    """
    A mixin for entities that have human-readable name or a label.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        Get the label.
        """
        pass


class Versioned(metaclass=abc.ABCMeta):
    """
    A mixin for entities that may have version.
    """

    @property
    @abc.abstractmethod
    def version(self) -> typing.Optional[str]:
        """
        Get a version `str` or `None` if the version is not available.
        """
        pass


class MetadataAware(metaclass=abc.ABCMeta):
    """
    A mixin for entities that have metadata.
    """

    @property
    @abc.abstractmethod
    def metadata(self) -> typing.MutableMapping[str, str]:
        """
        Get a mapping with entity metadata.
        """
        pass

    def metadata_to_str(self) -> str:
        """
        Dump the metadata to a `str`.
        """
        forbidden = {';', '='}
        for k, v in self.metadata.items():
            if any([token in k or token in v for token in forbidden]):
                raise ValueError(f'Metadata contains forbidden characters {forbidden}')

        return ';'.join([f'{k}={v}' for k, v in self.metadata.items()])

    @staticmethod
    def metadata_from_str(value: str) -> typing.Mapping[str, str]:
        """
        Load the metadata from `str` created by :py:func:`metadata_to_str`.
        """
        data = {}
        for item in value.split(';'):
            k, v = item.split('=')
            data[k] = v
        return data



import typing

from hpotk.model import TermId, Identified, ObservableFeature


# #################################################################################################################### #
# Non-public utilities
# #################################################################################################################### #


class SimpleFeature(Identified, ObservableFeature):

    def __init__(self, identifier: TermId, status: bool):
        self._id = identifier
        self._status = status

    @property
    def identifier(self) -> TermId:
        return self._id

    @identifier.setter
    def identifier(self, value: TermId):
        self._id = value

    @property
    def is_present(self) -> bool:
        return self._status


def map_to_stateful_feature(feature: typing.Union[Identified, TermId]) -> SimpleFeature:
    if isinstance(feature, Identified):
        term_id = feature.identifier
        if hasattr(feature, 'is_present'):
            attr = feature.is_present
            if hasattr(attr, '__call__'):
                status = attr()
                if not isinstance(status, bool):
                    raise ValueError(f'Feature {feature} has a callable `is_present` but it does not produce a `bool` '
                                     f'but {type(status)}')
            elif isinstance(attr, bool):
                status = attr
            else:
                raise ValueError(f'Feature {feature} has `is_present` attribute but the attribute is not a callable '
                                 f'or a `bool`')
        else:
            status = True  # we assume the feature was observed
    elif isinstance(feature, TermId):
        term_id = feature
        status = True  # we assume the feature was observed
    else:
        raise ValueError(f'Feature {feature} must implement `TermId` or `Identified`')

    return SimpleFeature(term_id, status)



"""
The `hpotk.validate` package provides code for Q/C of HPO applications.
"""

from ._model import ValidationLevel, ValidationResult, ValidationResults, RuleValidator, ValidationRunner
from ._hpo import AnnotationPropagationValidator, PhenotypicAbnormalityValidator, ObsoleteTermIdsValidator

__all__ = [
    'RuleValidator',
    'ValidationResult', 'ValidationLevel', 'ValidationResults',
    'AnnotationPropagationValidator', 'PhenotypicAbnormalityValidator', 'ObsoleteTermIdsValidator',
    'ValidationRunner'
]
